import logging

from django.apps import apps
from django.core.exceptions import ImproperlyConfigured
from django.utils.functional import classproperty

from mayan.apps.common.class_mixins import AppsModuleLoaderMixin
from mayan.apps.common.utils import get_class_full_name

from .class_mixins import DynamicFormBackendMixin

logger = logging.getLogger(name=__name__)


class BackendMetaclass(type):
    def __new__(mcs, name, bases, attrs):
        new_class = super().__new__(mcs, name, bases, attrs)

        base_backend_class = BackendMetaclass._get_base_backend_class(
            klass=new_class
        )

        _loader_module_name = getattr(new_class, '_loader_module_name')

        # Check if `_loader_module_name` is set as to not process the
        # metaclass.
        if _loader_module_name and base_backend_class:
            if base_backend_class != new_class:
                # Get this new child class full name, to be used as the
                # registry key.
                new_class_full_name = get_class_full_name(klass=new_class)
                if _loader_module_name in new_class_full_name:
                    # Only load classes defined in the loader module.
                    # This ensures miscellaneous classes like `NullBackends`
                    # are not registered.
                    base_backend_class.register(klass=new_class)

        return new_class

    @staticmethod
    def _get_base_backend_class(klass):
        # Filter class hierarchy. Select only the class with App loader
        # support. This means that is the class meant to work as the
        # app's backend class.
        base_backend_class = [
            parent for parent in klass.mro() if '_loader_module_name' in parent.__dict__ and parent._loader_module_name
        ]

        # No more than one app backend class should be returned.
        if len(base_backend_class) > 1:
            raise ImproperlyConfigured(
                'More than one backend parent class was returned. Should be '
                'only one.'
            )

        if base_backend_class:
            return base_backend_class[0]
        else:
            return None


class BaseBackend(AppsModuleLoaderMixin, metaclass=BackendMetaclass):
    """
    is_visible: Controls whether the backend will show up as a selection
                to the user.
    """
    _backend_identifier = 'backend_class_path'
    _registry = {}
    is_visible = True

    @classproperty
    def backend_class_path(cls):
        return cls.get_class_path()

    @classproperty
    def backend_id(cls):
        return getattr(cls, cls._backend_identifier)

    @classmethod
    def get(cls, name):
        return cls._registry.get(
            cls, {}
        )[name]

    @classmethod
    def get_all(cls):
        return list(
            cls._registry.get(
                cls, {}
            ).values()
        )

    @classmethod
    def get_choices(cls):
        choices = [
            (
                backend.backend_id, backend.label
            ) for backend in cls.get_all() if backend.is_visible
        ]

        choices.sort(
            key=lambda x: x[1]
        )

        return choices

    @classmethod
    def get_class_path(cls):
        return get_class_full_name(klass=cls)

    @classmethod
    def register(cls, klass):
        # Initialize the app backend class registry.
        cls._registry.setdefault(
            cls, {}
        )

        # Get this new child class full name, to be used as the
        # registry key.
        registry_key = klass.backend_id

        # Add the new child class to the app backend class registry.
        cls._registry[cls][registry_key] = klass


class ModelBaseBackend(BaseBackend):
    _backend_app_label = None
    _backend_model_name = None

    @classmethod
    def register(cls, klass):
        super().register(klass=klass)

    def get_model_instance(self):
        BackendModel = apps.get_model(
            app_label=self._backend_app_label,
            model_name=self._backend_model_name
        )

        return BackendModel.objects.get(pk=self.model_instance_id)

    def __init__(self, model_instance_id, **kwargs):
        self.model_instance_id = model_instance_id
        self.kwargs = kwargs


class StoredBaseBackend(BaseBackend):
    _backend_app_label = None
    _backend_model_name = None

    @classmethod
    def register(cls, klass):
        super().register(klass=klass)
        klass.get_model_instance()

    @classmethod
    def get_model_instance(cls):
        StoredBackendModel = apps.get_model(
            app_label=cls._backend_app_label,
            model_name=cls._backend_model_name
        )

        backend_path = get_class_full_name(klass=cls)

        instance, created = StoredBackendModel.objects.get_or_create(
            backend_path=backend_path
        )
        cls._backend_stored_model_id = instance.pk
        return instance

    def __init__(self, **kwargs):
        self.kwargs = kwargs


class DynamicFormModelBackend(DynamicFormBackendMixin, ModelBaseBackend):
    """
    Class that provides model stored backends which produce dynamic forms
    for data entry.
    """



import types

from django import forms as django_forms
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string


class FormMixinFilteredFieldsReload:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def do_fields_reload(self):
        # Updated filtered fields.
        field_reload_attributes = self.get_field_reload_attributes()

        for field_name in self.fields:
            field_instance = self.fields[field_name]
            if hasattr(field_instance, 'reload'):
                for key, value in field_reload_attributes.items():
                    setattr(field_instance, key, value)

                field_instance.reload()

    def get_field_reload_attributes(self):
        return {}


class FormMixinDynamicFields(FormMixinFilteredFieldsReload):
    def __init__(self, schema, *args, **kwargs):
        self.schema = schema

        super().__init__(*args, **kwargs)

        self.fieldsets = self.schema.get(
            'fieldsets', ()
        )
        self.fieldset_exclude_list = self.schema.get(
            'fieldset_exclude_list', ()
        )

        widgets = self.schema.get(
            'widgets', {}
        )

        field_order = self.get_field_order()

        for field_name in field_order:
            field_data = self.schema['fields'][field_name]
            field_class = import_string(
                dotted_path=field_data['class']
            )
            kwargs = {
                'label': field_data['label'],
                'required': field_data.get('required', True),
                'initial': field_data.get('default', None),
                'help_text': field_data.get('help_text')
            }

            widget = widgets.get(field_name)
            if widget:
                kwargs['widget'] = import_string(
                    dotted_path=widget['class']
                )(
                    **widget.get(
                        'kwargs', {}
                    )
                )

            kwargs.update(
                field_data.get(
                    'kwargs', {}
                )
            )
            self.fields[field_name] = field_class(**kwargs)

    def get_field_order(self):
        return self.schema.get(
            'field_order', self.schema['fields'].keys()
        )

    @property
    def media(self):
        """
        Append the media of the dynamic fields to the normal fields' media.
        """
        media = super().media
        media += django_forms.Media(
            **self.schema.get(
                'media', {}
            )
        )
        return media


class FormMixinFormMeta:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._form_meta = self._get_form_meta()

    def _get_form_meta(self, options_class_name='FormMeta'):
        merged = {}

        lineage = reversed(
            type(self).mro()
        )

        for base in lineage:
            options = getattr(base, options_class_name, None)
            if options:
                for name in dir(options):
                    if not name.startswith('_'):
                        merged[name] = getattr(options, name)

        return types.SimpleNamespace(**merged)


class FormMixinFieldsets(FormMixinFormMeta):
    fieldset_exclude_list = None
    fieldsets = None

    def get_fieldset_exclude_list(self):
        return self.fieldset_exclude_list or ()

    def get_fieldsets(self):
        fieldsets = getattr(self._form_meta, 'fieldsets', self.fieldsets)

        if fieldsets:
            fieldsets_field_list = []
            for fieldset, data in fieldsets:
                fieldsets_field_list.extend(
                    data['fields']
                )

            set_fields = set(self.fields)
            set_fieldsets = set(fieldsets_field_list)

            fieldset_exclude_list = self.get_fieldset_exclude_list()

            if fieldset_exclude_list:
                set_fields -= set(fieldset_exclude_list)
                set_fieldsets -= set(fieldset_exclude_list)

            if set_fields != set_fieldsets:
                raise ImproperlyConfigured(
                    'Mismatch fieldset fields: `{fields}` in form `{form}`'.format(
                        fields=', '.join(
                            set_fields.symmetric_difference(set_fieldsets)
                        ), form=self.__class__.__name__
                    )
                )

            return fieldsets
        else:
            return (
                (
                    None, {
                        'fields': tuple(self.fields)
                    }
                ),
            )



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0014_auto_20150708_0107'),
    ]

    operations = [
        migrations.AlterField(
            model_name='documenttype',
            name='label',
            field=models.CharField(
                unique=True, max_length=32, verbose_name='Label'
            ),
            preserve_default=True,
        ),
    ]



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('file_metadata', '0014_document_type_driver_populate')
    ]

    operations = [
        migrations.DeleteModel(name='DocumentTypeSettings')
    ]



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('file_caching', '0005_auto_20200322_0607')
    ]

    operations = [
        migrations.RemoveField(
            model_name='cache',
            name='label'
        ),
        migrations.RemoveField(
            model_name='cache',
            name='name'
        ),
        migrations.RemoveField(
            model_name='cache',
            name='storage_instance_path'
        ),
        migrations.AddField(
            model_name='cache',
            name='defined_storage_name',
            field=models.CharField(
                db_index=True, default='', help_text='Internal name of '
                'the defined storage for this cache.', max_length=96,
                unique=True, verbose_name='Defined storage name'
            ),
            preserve_default=False
        )
    ]



from django.template import Library

from ..literals import (
    USER_IMPERSONATE_VARIABLE_ID, USER_IMPERSONATE_VARIABLE_PERMANENT
)

register = Library()


@register.simple_tag(
    name='authentication_impersonation_check', takes_context=True
)
def tag_authentication_impersonation_check(context):
    request = getattr(context, 'request', None)

    if request:
        user_id = request.session.get(USER_IMPERSONATE_VARIABLE_ID)
        impersonate_permanent_session = USER_IMPERSONATE_VARIABLE_PERMANENT in request.session

        if user_id and not impersonate_permanent_session:
            return context.request.user
        else:
            return False
    else:
        return False



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.links import Link

from .icons import (
    icon_api, icon_api_documentation, icon_api_documentation_redoc
)
from .settings import setting_disable_links


def condition_api_links_enabled(context, resolved_object):
    return not setting_disable_links.value


link_api = Link(
    condition=condition_api_links_enabled, icon=icon_api, tags='new_window',
    text=_(message='REST API'), view='rest_api:api_root'
)
link_api_documentation = Link(
    condition=condition_api_links_enabled, icon=icon_api_documentation,
    tags='new_window', text=_(message='API Documentation (Swagger)'),
    view='rest_api:schema-swagger-ui'
)
link_api_documentation_redoc = Link(
    condition=condition_api_links_enabled, icon=icon_api_documentation_redoc,
    tags='new_window', text=_(message='API Documentation (ReDoc)'),
    view='rest_api:schema-redoc'
)



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('appearance', '0003_auto_20210823_2114')
    ]

    operations = [
        migrations.RemoveField(model_name='userthemesetting', name='theme'),
        migrations.RemoveField(model_name='userthemesetting', name='user'),
        migrations.DeleteModel(name='Theme'),
        migrations.DeleteModel(name='UserThemeSetting')
    ]



from mayan.apps.dependencies.classes import PythonDependency

PythonDependency(
    module=__name__, name='django-stronghold', version_string='==0.4.0'
)
PythonDependency(
    module=__name__, name='furl', version_string='==2.1.4'
)



from mayan.apps.icons.icons import Icon

icon_source_column_help_text = Icon(
    driver_name='fontawesome', symbol='question'
)
icon_source_column_help_text = Icon(
    driver_name='fontawesome-css', css_classes='far fa-question-circle'
)



from django.db import models
from django.utils.translation import gettext_lazy as _

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import (
    EventManagerMethodAfter, EventManagerSave
)

from ..events import event_metadata_type_relationship_updated
from ..managers import DocumentTypeMetadataTypeManager

from .metadata_type_models import MetadataType


class DocumentTypeMetadataType(ExtraDataModelMixin, models.Model):
    """
    Model used to store the relationship between a metadata type and a
    document type.
    """
    document_type = models.ForeignKey(
        on_delete=models.CASCADE, related_name='metadata', to=DocumentType,
        verbose_name=_(message='Document type')
    )
    metadata_type = models.ForeignKey(
        on_delete=models.CASCADE, related_name='document_types',
        to=MetadataType, verbose_name=_(message='Metadata type')
    )
    required = models.BooleanField(
        default=False, verbose_name=_(message='Required')
    )

    objects = DocumentTypeMetadataTypeManager()

    class Meta:
        ordering = ('metadata_type',)
        unique_together = ('document_type', 'metadata_type')
        verbose_name = _(message='Document type metadata type options')
        verbose_name_plural = _(
            message='Document type metadata types options'
        )

    def __str__(self):
        return str(self.metadata_type)

    @method_event(
        action_object='metadata_type',
        event=event_metadata_type_relationship_updated,
        event_manager_class=EventManagerMethodAfter,
        target='document_type'
    )
    def delete(self, *args, **kwargs):
        return super().delete(*args, **kwargs)

    @method_event(
        created={
            'action_object': 'metadata_type',
            'event': event_metadata_type_relationship_updated,
            'target': 'document_type'
        },
        edited={
            'action_object': 'metadata_type',
            'event': event_metadata_type_relationship_updated,
            'target': 'document_type'
        },
        event_manager_class=EventManagerSave
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)



from django.utils.translation import gettext_lazy as _

from mayan.apps.credentials.classes import CredentialBackend


class CredentialBackendGoogleServiceAccount(CredentialBackend):
    form_field_order = (
        'project_id', 'private_key_id', 'private_key', 'client_email',
        'client_id', 'auth_uri', 'token_uri', 'auth_provider_x509_cert_url',
        'client_x509_cert_url'
    )
    form_field_widgets = {
        'private_key': {
            'class': 'django.forms.widgets.Textarea'
        }
    }
    form_fields = {
        'project_id': {
            'label': _('Project ID'),
            'class': 'django.forms.CharField', 'default': '',
            'kwargs': {
                'max_length': 254
            }, 'required': True
        },
        'private_key_id': {
            'label': _('Private Key ID'),
            'class': 'django.forms.CharField', 'default': '',
            'kwargs': {
                'max_length': 254
            }, 'required': True
        },
        'private_key': {
            'label': _('Private Key'),
            'class': 'django.forms.CharField', 'default': ''
        },
        'client_email': {
            'label': _('Client email'),
            'class': 'django.forms.CharField', 'default': '',
            'kwargs': {
                'max_length': 254
            }, 'required': True
        },
        'client_id': {
            'label': _('Client ID'),
            'class': 'django.forms.CharField', 'default': '',
            'kwargs': {
                'max_length': 254
            }, 'required': True
        },
        'auth_uri': {
            'label': _('Authentication URI'),
            'class': 'django.forms.CharField',
            'default': 'https://accounts.google.com/o/oauth2/auth',
            'kwargs': {
                'max_length': 254
            }, 'required': True
        },
        'token_uri': {
            'label': _('Token URI'),
            'class': 'django.forms.CharField',
            'default': 'https://oauth2.googleapis.com/token',
            'kwargs': {
                'max_length': 254
            }, 'required': True
        },
        'auth_provider_x509_cert_url': {
            'label': _('X509 certificate provider URL'),
            'class': 'django.forms.CharField',
            'default': 'https://www.googleapis.com/oauth2/v1/certs',
            'kwargs': {
                'max_length': 254
            }, 'required': True
        },
        'client_x509_cert_url': {
            'label': _('Client X509 certificate URL'),
            'class': 'django.forms.CharField', 'default': '',
            'kwargs': {
                'max_length': 254
            }, 'required': True
        }
    }
    label = _('Google Service Account')

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _('Project'), {
                    'fields': ('project_id',)
                }
            ),
            (
                _('Secrets'), {
                    'fields': ('private_key_id', 'private_key', 'token_uri')
                }
            ),
            (
                _('Indentity'), {
                    'fields': ('client_email', 'client_id', 'auth_uri')
                }
            ),
            (
                _('Certificate'), {
                    'fields': (
                        'auth_provider_x509_cert_url', 'client_x509_cert_url'
                    )
                }
            )
        )

        return fieldsets

    @classmethod
    def post_processing(cls, obj):
        for key, value in obj.items():
            obj[key] = value.replace('\\n', '\n')

        return obj



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .literals import (
    DEFAULT_OCR_AUTO_OCR, DEFAULT_OCR_BACKEND, DEFAULT_OCR_BACKEND_ARGUMENTS
)
from .setting_migrations import OCRSettingMigration

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='OCR'), migration_class=OCRSettingMigration, name='ocr',
    version='0003'
)

setting_auto_ocr = setting_namespace.do_setting_add(
    choices=('false', 'true'), default=DEFAULT_OCR_AUTO_OCR,
    global_name='OCR_AUTO_OCR', help_text=_(
        message='Set new document types to perform OCR automatically by default.'
    )
)
setting_ocr_backend = setting_namespace.do_setting_add(
    default=DEFAULT_OCR_BACKEND, global_name='OCR_BACKEND', help_text=_(
        message='Full path to the backend to be used to do OCR.'
    )
)
setting_ocr_backend_arguments = setting_namespace.do_setting_add(
    default=DEFAULT_OCR_BACKEND_ARGUMENTS,
    global_name='OCR_BACKEND_ARGUMENTS', help_text=_(
        message='Arguments to pass to `OCR_BACKEND`.'
    )
)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('file_caching', '0006_auto_20200322_0626')
    ]

    operations = [
        migrations.AddField(
            model_name='cachepartitionfile', name='hits',
            field=models.PositiveIntegerField(
                default=0, help_text='Times this cache partition file has '
                'been accessed.', verbose_name='Hits'
            )
        )
    ]



import json

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import F, Value
from django.utils.timezone import now

from mayan.apps.databases.manager_mixins import ManagerMinixCreateBulk

from .settings import (
    setting_saved_resultset_results_limit,
    setting_saved_resultset_time_to_live,
    setting_saved_resultsets_per_user_limit
)


class SavedResultsetEntryManager(ManagerMinixCreateBulk, models.Manager):
    """
    Nothing additional required, this is just to add the create bulk mixing
    to the manager.
    """


class SavedResultsetManager(models.Manager):
    def expired_delete(self):
        datetime_current = now()
        epoch_end = int(
            datetime_current.timestamp()
        )

        queryset = self.annotate(
            epoch_diff=Value(epoch_end) - F('epoch')
        )

        queryset = queryset.filter(
            epoch_diff__gt=F('time_to_live')
        )

        queryset.delete()

    def queryset_save(
        self, queryset, search_explainer_text, search_query, user
    ):
        SavedResultsetEntry = apps.get_model(
            app_label='dynamic_search', model_name='SavedResultsetEntry'
        )

        queryset_per_user = self.filter(user=user).only('id')

        queryset_per_user_to_delete = queryset_per_user[
            setting_saved_resultsets_per_user_limit.value:
        ]

        self.filter(pk__in=queryset_per_user_to_delete).delete()

        content_type = ContentType.objects.get_for_model(model=queryset.model)

        queryset_results = queryset.values_list('pk', flat=True)[
            :setting_saved_resultset_results_limit.value
        ]

        search_query_text = json.dumps(obj=search_query)
        datetime_current = now()
        saved_resultset = self.model(
            app_label=content_type.app_label,
            epoch=datetime_current.timestamp(), model_name=content_type.model,
            result_count=queryset_results.count(),
            search_explainer_text=search_explainer_text,
            search_query_text=search_query_text,
            time_to_live=setting_saved_resultset_time_to_live.value,
            timestamp=datetime_current, user=user
        )
        saved_resultset._event_actor = user
        saved_resultset.save()

        coroutine = SavedResultsetEntry.objects.create_bulk()
        next(coroutine)

        for pk in queryset_results.iterator():
            coroutine.send(
                {'object_id': pk, 'saved_resultset_id': saved_resultset.pk}
            )
        coroutine.close()

        return saved_resultset



import logging

from django.core.files.base import File

from mayan.apps.mime_types.classes import MIMETypeBackend
from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import (
    ExternalContentTypeObjectAPIViewMixin
)

from .api_view_mixins import APIImageViewMixin
from .classes import AppImageErrorImage
from .models import Asset
from .permissions import (
    permission_asset_create, permission_asset_delete, permission_asset_edit,
    permission_asset_view
)
from .serializers import AppImageErrorSerializer, AssetSerializer

logger = logging.getLogger(name=__name__)


class APIAssetListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the assets.
    post: Create a new asset.
    """
    mayan_object_permission_map = {'GET': permission_asset_view}
    mayan_view_permission_map = {'POST': permission_asset_create}
    serializer_class = AssetSerializer
    source_queryset = Asset.objects.all()

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class APIAssetDetailView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete: Delete the selected asset.
    get: Return the details of the selected asset.
    patch: Edit the properties of the selected asset.
    put: Edit the properties of the selected asset.
    """
    lookup_url_kwarg = 'asset_id'
    mayan_object_permission_map = {
        'DELETE': permission_asset_delete,
        'GET': permission_asset_view,
        'PATCH': permission_asset_edit,
        'PUT': permission_asset_edit
    }
    serializer_class = AssetSerializer
    source_queryset = Asset.objects.all()

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class APIAssetImageView(APIImageViewMixin, generics.RetrieveAPIView):
    """
    get: Returns an image representation of the selected asset.
    """
    lookup_url_kwarg = 'asset_id'
    mayan_object_permission_map = {'GET': permission_asset_view}
    source_queryset = Asset.objects.all()


class APIAppImageErrorDetailView(generics.RetrieveAPIView):
    """
    get: Returns the details of the selected app image error.
    """
    lookup_url_kwarg = 'app_image_error_name'
    serializer_class = AppImageErrorSerializer

    def get_object(self):
        return AppImageErrorImage.get(
            name=self.kwargs[self.lookup_url_kwarg]
        )


class APIAppImageErrorImageView(APIImageViewMixin, generics.RetrieveAPIView):
    """
    get: Returns an image representation of the selected app image error.
    """
    lookup_url_kwarg = 'app_image_error_name'

    def get_file_generator(self):
        def file_generator():
            with self.get_object().open() as file_object:
                while True:
                    chunk = file_object.read(File.DEFAULT_CHUNK_SIZE)
                    if not chunk:
                        break
                    else:
                        yield chunk

        return file_generator

    def get_object(self):
        return AppImageErrorImage.get(
            name=self.kwargs[self.lookup_url_kwarg]
        )

    def get_stream_mime_type(self):
        mime_type_backend = MIMETypeBackend.get_backend_instance()
        with self.get_object().open() as file_object:
            mime_type, mime_encoding = mime_type_backend.get_mime_type(
                file_object=file_object, mime_type_only=True
            )
            return mime_type

    def set_cache_file(self, request):
        return


class APIAppImageErrorListView(generics.ListAPIView):
    """
    get: Returns an list of the available app image errors.
    """
    serializer_class = AppImageErrorSerializer

    def get_source_queryset(self):
        return list(
            AppImageErrorImage.all()
        )


class APIContentObjectImageView(
    APIImageViewMixin, ExternalContentTypeObjectAPIViewMixin,
    generics.RetrieveAPIView
):
    """
    get: Returns an image representation of the selected content object.
    """
    def get_content_type(self):
        return self.content_type

    def set_object(self):
        self.obj = self.get_external_object()



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Statistics'), name='statistics'
)

permission_statistics_view = namespace.add_permission(
    label=_(message='View statistics'), name='statistics_view'
)



from django.apps import apps
from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
from django.db import models
from django.db.models.constants import LOOKUP_SEP
from django.utils.translation import gettext_lazy as _

from .utils import help_text_for_field_recursive


class ModelAttribute:
    _class_registry = []
    _model_registry = {}

    @classmethod
    def get_all_choices_for(cls, model):
        result = []

        for klass in cls._class_registry:
            klass_choice_list = klass.get_choices_for(model=model)
            if klass_choice_list:
                result.append(
                    (klass.class_label, klass_choice_list)
                )

        return result

    @classmethod
    def get_choices_for(cls, model):
        klass_choice_list = cls.get_for(model=model)

        return sorted(
            (
                (
                    entry.name, entry.get_display()
                ) for entry in klass_choice_list
            ), key=lambda x: x[1]
        )

    @classmethod
    def get_for(cls, model):
        try:
            return cls._model_registry[cls.class_name][model]
        except KeyError:
            # We were passed a model instance, try again using the model of
            # the instance.

            # If we are already in the model class, exit with an error.
            if model.__class__ == models.base.ModelBase:
                return []

            return cls.get_for(
                model=type(model)
            )

    @classmethod
    def register(cls, klass):
        cls._class_registry.append(klass)

    def __init__(self, model, name, label=None, description=None):
        self.model = model
        self.label = label
        self.name = name
        self.description = description
        self._model_registry.setdefault(
            self.class_name, {}
        )
        self._model_registry[self.class_name].setdefault(
            model, []
        )
        self._model_registry[self.class_name][model].append(self)

    def get_display(self, show_name=False):
        if self.description:
            template = '{label} - {description}'
        else:
            template = '{label}'

        label = self.name if show_name else self.label

        result = template.format(label=label, description=self.description)

        return result

    def get_description(self):
        return self.description

    get_description.short_description = _('Description')

    def get_label(self):
        return self.label

    get_label.short_description = _('Label')

    def get_name(self):
        return self.name

    get_name.help_text = _('Example usage of the property.')
    get_name.short_description = _('Name')


class ModelField(ModelAttribute):
    class_label = _(message='Model fields')
    class_name = 'field'

    def __init__(self, **kwargs):
        if 'label' in kwargs:
            raise ImproperlyConfigured(
                '`ModelField` and subclasses not longer accept `label` '
                'argument. Ensure the `verbose_name` is set in the model '
                'instead.'
            )

        super().__init__(**kwargs)

        self.do_description_set()
        self.do_label_set()

    def do_description_set(self):
        if not self.description:
            self.description = help_text_for_field_recursive(
                model=self.model, name=self.name
            )

    def do_label_set(self):
        text_model_list = []
        last_model = self.model
        for part in self.name.split(LOOKUP_SEP):
            try:
                field = last_model._meta.get_field(field_name=part)
            except FieldDoesNotExist:
                break
            else:
                verbose_name = str(last_model._meta.verbose_name)
                if last_model != self.model and verbose_name:
                    text_model_list.append(
                        '{}'.format(verbose_name)
                    )

                last_model = field.related_model or field.model

        text_model_list.append(
            str(field.verbose_name)
        )

        self.label = ' > '.join(text_model_list)


class ModelFieldRelated(ModelField):
    class_label = _(message='Model related fields')
    class_name = 'related_field'


class ModelProperty(ModelAttribute):
    class_label = _(message='Model properties')
    class_name = 'property'


class ModelReverseField(ModelField):
    class_label = _(message='Model reverse fields')
    class_name = 'reverse_field'

    def __init__(self, *args, **kwargs):
        super(ModelField, self).__init__(*args, **kwargs)
        self._final_model_verbose_name = None

        if not self.label:
            self.label = self.get_field_attribute(
                attribute='verbose_name_plural'
            )

    def get_field_attribute(self, attribute):
        field = self.model._meta.get_field(field_name=self.name)

        return getattr(field.related_model._meta, attribute)


class ModelQueryFields:
    _registry = {}

    @classmethod
    def get(cls, model):
        try:
            return cls._registry[model]
        except KeyError:
            ModelQueryFields(model=model)
            return cls.get(model=model)

    def __init__(self, model):
        self.model = model
        self.select_related_fields = []
        self.prefetch_related_fields = []
        self.__class__._registry[model] = self

    def add_select_related_field(self, field_name):
        if field_name in self.select_related_fields:
            raise ImproperlyConfigured(
                '"{}" model already has a "{}" query select '
                'related field.'.format(
                    self.model, field_name
                )
            )
        self.select_related_fields.append(field_name)

    def add_prefetch_related_field(self, field_name):
        if field_name in self.prefetch_related_fields:
            raise ImproperlyConfigured(
                '"{}" model already has a "{}" query prefetch '
                'related field.'.format(
                    self.model, field_name
                )
            )
        self.prefetch_related_fields.append(field_name)

    def get_queryset(self, manager_name=None):
        if manager_name:
            manager = getattr(self.model, manager_name)
        else:
            manager = self.model._meta.default_manager

        queryset = manager.all()

        if self.select_related_fields:
            queryset = queryset.select_related(*self.select_related_fields)

        if self.prefetch_related_fields:
            queryset = queryset.prefetch_related(
                *self.prefetch_related_fields
            )

        return queryset


class ModelWrapper:
    @classmethod
    def all(cls):
        class_registry = ModelProperty._model_registry[
            ModelProperty.class_name
        ]

        result = []

        for entry in class_registry.keys():
            instance = cls(model=entry)
            result.append(instance)

        result.sort(
            key=lambda entry: entry.get_name_full()
        )

        return result

    def __init__(self, model):
        self.model = model

        meta = self.model._meta

        self.app_config = meta.app_config
        self.model_name = meta.model_name
        self.verbose_name = meta.verbose_name

    def get_app_label(self):
        return self.app_config.verbose_name

    get_app_label.help_text = _('Name of the app where the model is defined.')
    get_app_label.short_description = _('App')

    def get_name_full(self):
        return '{}.{}'.format(self.app_config.label, self.model_name)

    get_name_full.help_text = _('Complete unique name of the model.')
    get_name_full.short_description = _('Full name')

    def get_label(self):
        return self.verbose_name

    get_label.help_text = _('Human readable name of the model.')
    get_label.short_description = _('Label')

    def get_name(self):
        return self.model_name

    get_name.short_description = _('Name')


class QuerysetParametersSerializer:
    @staticmethod
    def decompose(_model, _method_name, _manager_name=None, **kwargs):
        ContentType = apps.get_model(
            app_label='contenttypes', model_name='ContentType'
        )

        _manager_name = _manager_name or _model._meta.default_manager.name

        serialized_kwargs = []
        for name, value in kwargs.items():
            try:
                content_type = ContentType.objects.get_for_model(model=value)
            except AttributeError:
                """The value is not a model instance, pass it as-is."""
                serialized_kwargs.append(
                    {
                        'name': name,
                        'value': value
                    }
                )
            else:
                serialized_kwargs.append(
                    {
                        'name': name,
                        'content_type_id': content_type.pk,
                        'object_id': value.pk
                    }
                )

        return {
            'model_content_type_id': ContentType.objects.get_for_model(
                model=_model
            ).pk,
            'manager_name': _manager_name,
            'method_name': _method_name,
            'kwargs': serialized_kwargs
        }

    @staticmethod
    def rebuild(decomposed_queryset):
        ContentType = apps.get_model(
            app_label='contenttypes', model_name='ContentType'
        )

        model = ContentType.objects.get(
            pk=decomposed_queryset['model_content_type_id']
        ).model_class()

        queryset = getattr(
            model, decomposed_queryset['manager_name']
        )

        kwargs = {}

        parameter_list = decomposed_queryset.get(
            'kwargs', ()
        )

        for parameter in parameter_list:
            if 'content_type_id' in parameter:
                content_type = ContentType.objects.get(
                    pk=parameter['content_type_id']
                )
                value = content_type.get_object_for_this_type(
                    pk=parameter['object_id']
                )
            else:
                value = parameter['value']

            kwargs[
                parameter['name']
            ] = value

        queryset_method = getattr(
            queryset, decomposed_queryset['method_name']
        )
        return queryset_method(**kwargs)


ModelAttribute.register(klass=ModelProperty)
ModelAttribute.register(klass=ModelField)
ModelAttribute.register(klass=ModelFieldRelated)
ModelAttribute.register(klass=ModelReverseField)



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('events', '0008_auto_20180315_0029'),
        (
            'document_states',
            '0024_remove_duplicate_workflow_template_transition_triggers'
        )
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='workflowtransitiontriggerevent',
            unique_together={
                ('transition', 'event_type')
            }
        )
    ]



from django.db import models
from django.urls import reverse
from django.utils.translation import gettext, gettext_lazy as _

from mptt.fields import TreeForeignKey
from mptt.models import MPTTModel

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import (
    EventManagerMethodAfter, EventManagerSave
)
from mayan.apps.events.models import StoredEventType

from ..events import event_index_template_created, event_index_template_edited
from ..managers import IndexTemplateManager

from .index_template_model_mixins import (
    IndexTemplateBusinessLogicMixin, IndexTemplateNodeBusinessLogicMixin
)


class IndexTemplate(
    ExtraDataModelMixin, IndexTemplateBusinessLogicMixin, models.Model
):
    """
    Parent model that defines an index and hold all the relationship for its
    template and instance when resolved.
    """
    _ordering_fields = ('enabled', 'label', 'slug')

    label = models.CharField(
        help_text=_(message='Short description of this index.'),
        max_length=128, unique=True, verbose_name=_(message='Label')
    )
    slug = models.SlugField(
        help_text=_(
            message='This value will be used by other apps to reference this index.'
        ), max_length=128, unique=True, verbose_name=_(message='Slug')
    )
    enabled = models.BooleanField(
        default=True, help_text=_(
            message='Causes this index to be visible and updated when document data '
            'changes.'
        ), verbose_name=_(message='Enabled')
    )
    document_types = models.ManyToManyField(
        related_name='index_templates', to=DocumentType,
        verbose_name=_(message='Document types')
    )

    objects = IndexTemplateManager()

    class Meta:
        ordering = ('label',)
        verbose_name = _(message='Index template')
        verbose_name_plural = _(message='Index templates')

    def __str__(self):
        return self.label

    def get_absolute_url(self):
        return reverse(
            kwargs={'index_template_id': self.pk},
            viewname='indexing:index_template_view'
        )

    def natural_key(self):
        return (self.slug,)

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_index_template_created,
            'target': 'self',
        },
        edited={
            'event': event_index_template_edited,
            'target': 'self',
        }
    )
    def save(self, *args, **kwargs):
        is_new = not self.pk
        super().save(*args, **kwargs)
        if is_new:
            # Automatically create the root index template node.
            IndexTemplateNode.objects.get_or_create(parent=None, index=self)
            self.do_event_triggers_populate()

        self.index_template_root_node.initialize_index_instance_root_node()


class IndexTemplateEventTrigger(ExtraDataModelMixin, models.Model):
    index_template = models.ForeignKey(
        on_delete=models.CASCADE, related_name='event_triggers',
        to=IndexTemplate, verbose_name=_(message='Index template')
    )
    stored_event_type = models.ForeignKey(
        on_delete=models.CASCADE, to=StoredEventType,
        verbose_name=_(message='Event type')
    )

    class Meta:
        unique_together = ('index_template', 'stored_event_type')
        verbose_name = _(message='Index template event trigger')
        verbose_name_plural = _(message='Index template event triggers')

    def __str__(self):
        return str(self.stored_event_type)

    @method_event(
        event_manager_class=EventManagerMethodAfter,
        event=event_index_template_edited,
        target='index_template'
    )
    def delete(self, *args, **kwargs):
        return super().delete(*args, **kwargs)

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_index_template_edited,
            'target': 'index_template'
        },
        edited={
            'event': event_index_template_edited,
            'target': 'index_template'
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)


class IndexTemplateNode(IndexTemplateNodeBusinessLogicMixin, MPTTModel):
    """
    The template to generate an index. Each entry represents a level in a
    hierarchy of levels. Each level can contain further levels or a list of
    documents but not both.
    """
    _ordering_fields = ('enabled', 'link_documents')

    parent = TreeForeignKey(
        blank=True, help_text=_(
            message='Parent index template node of this node.'
        ), null=True, on_delete=models.CASCADE, related_name='children',
        to='self'
    )
    index = models.ForeignKey(
        on_delete=models.CASCADE, related_name='index_template_nodes',
        to=IndexTemplate, verbose_name=_(message='Index')
    )
    expression = models.TextField(
        help_text=_(
            message='Enter a template to render. Use Django\'s default '
            'templating language.'
        ),
        verbose_name=_(message='Indexing expression')
    )
    enabled = models.BooleanField(
        default=True,
        help_text=_(
            message='Causes this node to be visible and updated when '
            'document data changes.'
        ),
        verbose_name=_(message='Enabled')
    )
    link_documents = models.BooleanField(
        default=False,
        help_text=_(
            message='Check this option to have this node act as a container '
            'for documents and not as a parent for further nodes.'
        ),
        verbose_name=_(message='Link documents')
    )

    class Meta:
        verbose_name = _(message='Index template node')
        verbose_name_plural = _(message='Index template nodes')

    def __str__(self):
        if self.is_root_node():
            return gettext(message='Root')
        else:
            return self.expression



from django.utils.translation import gettext_lazy as _

from mayan.apps.app_manager.apps import MayanAppConfig


class FormsApp(MayanAppConfig):
    app_namespace = 'forms'
    app_url = 'forms'
    has_javascript_translations = True
    has_static_media = True
    label = 'mayan_forms'  # Avoid clash with Django's `forms` app.
    name = 'mayan.apps.forms'
    static_media_ignore_patterns = (
        'mayan_forms/node_modules/dropzone/index.js',
        'mayan_forms/node_modules/dropzone/component.json'
    )
    verbose_name = _(message='Forms')



from django.contrib.auth.views import (
    PasswordResetCompleteView, PasswordResetConfirmView,
    PasswordResetDoneView, PasswordResetView
)
from django.shortcuts import redirect
from django.urls import reverse_lazy

from stronghold.views import StrongholdPublicMixin

import mayan
from mayan.apps.common.settings import setting_home_view
from mayan.apps.organizations.utils import get_organization_installation_url

from ..settings import setting_disable_password_reset


class MayanPasswordResetRedirectMixin:
    def get(self, *args, **kwargs):
        if setting_disable_password_reset.value:
            return redirect(to=setting_home_view.value)
        return super().get(*args, **kwargs)

    def post(self, *args, **kwargs):
        if setting_disable_password_reset.value:
            return redirect(to=setting_home_view.value)
        return super().post(*args, **kwargs)


class MayanPasswordResetCompleteView(
    MayanPasswordResetRedirectMixin, StrongholdPublicMixin,
    PasswordResetCompleteView
):
    extra_context = {
        'appearance_type': 'plain'
    }
    template_name = 'authentication/password_reset_complete.html'


class MayanPasswordResetConfirmView(
    MayanPasswordResetRedirectMixin, StrongholdPublicMixin,
    PasswordResetConfirmView
):
    extra_context = {
        'appearance_type': 'plain'
    }
    success_url = reverse_lazy(
        viewname='authentication:password_reset_complete_view'
    )
    template_name = 'authentication/password_reset_confirm.html'


class MayanPasswordResetDoneView(
    MayanPasswordResetRedirectMixin, StrongholdPublicMixin,
    PasswordResetDoneView
):
    extra_context = {
        'appearance_type': 'plain'
    }
    template_name = 'authentication/password_reset_done.html'


class MayanPasswordResetView(
    MayanPasswordResetRedirectMixin, StrongholdPublicMixin, PasswordResetView
):
    email_template_name = 'authentication/password_reset_email.html'
    extra_context = {
        'appearance_type': 'plain'
    }
    subject_template_name = 'authentication/password_reset_subject.txt'
    success_url = reverse_lazy(
        viewname='authentication:password_reset_done_view'
    )
    template_name = 'authentication/password_reset_form.html'

    # Hardcoded overloaded method to allow adding extra email context from
    # a method and not just the Django provided `self.extra_email_context`.
    # On each new Django version, verify if this method has changed and
    # update this overloading.
    def form_valid(self, form):
        opts = {
            'email_template_name': self.email_template_name,
            'extra_email_context': self.get_extra_email_context(),
            'from_email': self.from_email,
            'html_email_template_name': self.html_email_template_name,
            'request': self.request,
            'subject_template_name': self.subject_template_name,
            'token_generator': self.token_generator,
            'use_https': self.request.is_secure()
        }
        form.save(**opts)
        # Specify the super class `PasswordResetView` explicitly to avoid
        # executing `form_valid` again.
        return super(PasswordResetView, self).form_valid(form=form)

    def get_extra_email_context(self):
        extra_email_context_project_website = get_organization_installation_url(
            request=self.request
        )
        extra_email_context = {
            'project_copyright': mayan.__copyright__,
            'project_license': mayan.__license__,
            'project_title': mayan.__title__,
            'project_website': str(extra_email_context_project_website)
        }
        return extra_email_context



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.links import Link
from mayan.apps.navigation.utils import factory_condition_queryset_access

from .icons import (
    icon_announcement_create, icon_announcement_delete_multiple,
    icon_announcement_delete_single, icon_announcement_edit,
    icon_announcement_list
)
from .permissions import (
    permission_announcement_create, permission_announcement_delete,
    permission_announcement_edit, permission_announcement_view
)

link_announcement_create = Link(
    icon=icon_announcement_create, permission=permission_announcement_create,
    text=_(message='Create announcement'),
    view='announcements:announcement_create'
)
link_announcement_delete_multiple = Link(
    icon=icon_announcement_delete_multiple, tags='dangerous',
    text=_(message='Delete'),
    view='announcements:announcement_multiple_delete'
)
link_announcement_delete_single = Link(
    args='object.pk', icon=icon_announcement_delete_single,
    permission=permission_announcement_delete,
    tags='dangerous', text=_(message='Delete'),
    view='announcements:announcement_single_delete'
)
link_announcement_edit = Link(
    args='object.pk', icon=icon_announcement_edit,
    permission=permission_announcement_edit, text=_(message='Edit'),
    view='announcements:announcement_edit'
)
link_announcement_list = Link(
    icon=icon_announcement_list, text=_(message='Announcements'),
    view='announcements:announcement_list'
)
link_announcement_setup = Link(
    condition=factory_condition_queryset_access(
        app_label='announcements', model_name='Announcement',
        object_permission=permission_announcement_view,
        view_permission=permission_announcement_create,
    ), icon=icon_announcement_list,
    text=_(message='Announcements'), view='announcements:announcement_list'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.app_manager.apps import MayanAppConfig

from .literals import LIST_MODE_CHOICES


class ModelMixinAppConfig:
    def get_app_config(self):
        app_config = MayanAppConfig.get_by_namespace(namespace=self.namespace)

        return app_config

    get_app_config.help_text = _(message='The app to which the view belongs.')
    get_app_config.short_description = _(message='App')

    def get_app_config_verbose_name(self):
        app_config = self.get_app_config()

        return app_config.verbose_name

    get_app_config_verbose_name.help_text = _(
        message='The app to which the view belongs.'
    )
    get_app_config_verbose_name.short_description = _(message='App')


class ModelMixinUserConfirmViewBusinessLogic(ModelMixinAppConfig):
    """
    No other code besides the `ModelMixinAppConfig` inheritance.
    """


class ModelMixinUserViewModeBusinessLogic(ModelMixinAppConfig):
    def get_value_display(self):
        map_list_mode_choices = dict(LIST_MODE_CHOICES)

        try:
            return map_list_mode_choices[self.value]
        except KeyError:
            return _(message='Unknown mode: %s') % self.value

    get_value_display.help_text = _(message='The current mode of the view.')
    get_value_display.short_description = _(message='Value')



from mayan.apps.icons.icons import Icon

icon_model_property_list = Icon(
    driver_name='fontawesome', symbol='code-commit'
)
icon_property_model_list = Icon(
    driver_name='fontawesome', symbol='window-restore'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.links import Link

from ..icons import (
    icon_document_list, icon_document_preview,
    icon_document_properties_detail, icon_document_properties_edit,
    icon_document_recently_accessed_list, icon_document_recently_created_list,
    icon_document_type_change_multiple, icon_document_type_change_single
)
from ..permissions import (
    permission_document_properties_edit, permission_document_view
)

link_document_list = Link(
    icon=icon_document_list,
    text=_(message='All documents'), view='documents:document_list'
)
link_document_preview = Link(
    args='resolved_object.id', icon=icon_document_preview,
    permission=permission_document_view, text=_(message='Preview'),
    view='documents:document_preview'
)
link_document_properties_detail = Link(
    args='resolved_object.id', icon=icon_document_properties_detail,
    permission=permission_document_view, text=_(message='Properties'),
    view='documents:document_properties'
)
link_document_properties_edit = Link(
    args='resolved_object.id',
    icon=icon_document_properties_edit,
    permission=permission_document_properties_edit,
    text=_(message='Edit properties'),
    view='documents:document_properties_edit'
)

# Document type

link_document_type_change_multiple = Link(
    text=_(message='Change type'), icon=icon_document_type_change_multiple,
    view='documents:document_multiple_type_change'
)
link_document_type_change_single = Link(
    args='resolved_object.id', icon=icon_document_type_change_single,
    permission=permission_document_properties_edit, text=_(
        message='Change type'
    ), view='documents:document_type_change'
)

# Recently accessed

link_document_recently_accessed_list = Link(
    icon=icon_document_recently_accessed_list, text=_(
        message='Recently accessed'
    ), view='documents:document_recently_accessed_list'
)

# Recently created

link_document_recently_created_list = Link(
    icon=icon_document_recently_created_list, text=_(
        message='Recently created'
    ), view='documents:document_recently_created_list'
)



import logging

from django.contrib import messages
from django.http import HttpResponseRedirect
from django.template import RequestContext
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from mayan.apps.views.generics import (
    FormView, MultipleObjectDeleteView, SingleObjectCreateView,
    SingleObjectEditView, SingleObjectListView
)
from mayan.apps.views.view_mixins import ExternalContentTypeObjectViewMixin

from ..forms import LayerTransformationSelectForm
from ..icons import (
    icon_transformation_create, icon_transformation_delete,
    icon_transformation_edit, icon_transformation_list,
    icon_transformation_select
)
from ..links import link_transformation_select
from ..models import ObjectLayer
from ..transformations import BaseTransformation

from .view_mixins import (
    ViewMixinDynamicTransformationFormClass, ViewMixinLayer,
    ViewMixinTransformationTemplateName
)

logger = logging.getLogger(name=__name__)


class TransformationCreateView(
    ViewMixinDynamicTransformationFormClass, ViewMixinLayer,
    ExternalContentTypeObjectViewMixin, ViewMixinTransformationTemplateName,
    SingleObjectCreateView
):
    view_icon = icon_transformation_create

    def form_valid(self, form):
        object_layer, created = ObjectLayer.objects.get_for(
            layer=self.layer, obj=self.external_object
        )

        instance = form.save(commit=False)
        instance.content_object = self.external_object
        instance.name = self.kwargs['transformation_name']
        instance.object_layer = object_layer

        try:
            instance.full_clean()
            instance.save()
        except Exception as exception:
            logger.debug('Invalid form, exception: %s', exception)
            messages.error(
                message=_(
                    message='Error creating transformation: %s.'
                ) % exception, request=self.request
            )
            return super().form_invalid(form=form)
        else:
            return super().form_valid(form=form)

    def get_extra_context(self):
        transformation_template_name = self.get_transformation_template_name()

        if transformation_template_name:
            form_field_css_classes = 'hidden'
        else:
            form_field_css_classes = ''

        return {
            'content_object': self.external_object,
            'form_field_css_classes': form_field_css_classes,
            'layer': self.layer,
            'layer_name': self.layer.name,
            'navigation_object_list': ('content_object',),
            'title': _(
                message='Create layer "%(layer)s" transformation '
                '"%(transformation)s" for: %(object)s'
            ) % {
                'layer': self.layer,
                'transformation': self.get_transformation_class(),
                'object': self.external_object
            }
        }

    def get_form_extra_kwargs(self):
        return {
            'initial': {'order': None},
            'transformation_name': self.kwargs['transformation_name']
        }

    def get_external_object_permission(self):
        return self.layer.get_permission(action='create')

    def get_post_action_redirect(self):
        return reverse(
            kwargs={
                'app_label': self.kwargs['app_label'],
                'model_name': self.kwargs['model_name'],
                'object_id': self.kwargs['object_id'],
                'layer_name': self.kwargs['layer_name']
            }, viewname='converter:transformation_list'
        )

    def get_transformation_class(self):
        return BaseTransformation.get(
            name=self.kwargs['transformation_name']
        )


class TransformationDeleteView(
    ViewMixinLayer, ExternalContentTypeObjectViewMixin,
    MultipleObjectDeleteView
):
    pk_url_kwarg = 'transformation_id'
    success_message_plural = _(message='%(count)d transformations deleted successfully.')
    success_message_single = _(
        message='Transformation "%(object)s" deleted successfully.'
    )
    success_message_singular = _(
        message='%(count)d transformation deleted successfully.'
    )
    title_plural = _(message='Delete the %(count)d selected transformations')
    title_single = _(message='Delete transformation: %(object)s')
    title_singular = _(message='Delete the %(count)d selected transformation')
    view_icon = icon_transformation_delete

    def get_external_object_permission(self):
        return self.layer.get_permission(action='delete')

    def get_extra_context(self):
        return {
            'content_object': self.external_object,
            'layer': self.layer,
            'layer_name': self.layer.name,
            'navigation_object_list': ('content_object', 'transformation'),
            'previous': self.get_post_action_redirect(),
        }

    def get_post_action_redirect(self):
        obj = self.object_list.first()

        return reverse(
            kwargs={
                'app_label': obj.object_layer.content_type.app_label,
                'model_name': obj.object_layer.content_type.model,
                'object_id': obj.object_layer.object_id,
                'layer_name': obj.object_layer.stored_layer.name
            }, viewname='converter:transformation_list'
        )

    def get_source_queryset(self):
        return self.layer.get_transformations_for(
            obj=self.external_object
        )


class TransformationEditView(
    ViewMixinLayer, ViewMixinDynamicTransformationFormClass,
    ExternalContentTypeObjectViewMixin, ViewMixinTransformationTemplateName,
    SingleObjectEditView
):
    pk_url_kwarg = 'transformation_id'
    view_icon = icon_transformation_edit

    def form_valid(self, form):
        instance = form.save(commit=False)
        try:
            instance.full_clean()
            instance.save()
        except Exception as exception:
            logger.debug('Invalid form, exception: %s', exception)
            return super().form_invalid(form=form)
        else:
            return super().form_valid(form=form)

    def get_external_object_permission(self):
        return self.layer.get_permission(action='edit')

    def get_extra_context(self):
        transformation_template_name = self.get_transformation_template_name()

        if transformation_template_name:
            form_field_css_classes = 'hidden'
        else:
            form_field_css_classes = ''

        return {
            'content_object': self.external_object,
            'form_field_css_classes': form_field_css_classes,
            'layer': self.layer,
            'layer_name': self.layer.name,
            'navigation_object_list': ('content_object', 'transformation'),
            'title': _(
                message='Edit transformation "%(transformation)s" '
                'for: %(content_object)s'
            ) % {
                'transformation': self.object,
                'content_object': self.external_object
            },
            'transformation': self.object
        }

    def get_post_action_redirect(self):
        return reverse(
            kwargs={
                'app_label': self.object.object_layer.content_type.app_label,
                'model_name': self.object.object_layer.content_type.model,
                'object_id': self.object.object_layer.object_id,
                'layer_name': self.object.object_layer.stored_layer.name
            }, viewname='converter:transformation_list'
        )

    def get_source_queryset(self):
        return self.layer.get_transformations_for(
            obj=self.external_object
        )

    def get_transformation_class(self):
        return self.object.get_transformation_class()


class TransformationListView(
    ViewMixinLayer, ExternalContentTypeObjectViewMixin, SingleObjectListView
):
    view_icon = icon_transformation_list

    def get_external_object_permission(self):
        return self.layer.get_permission(action='view')

    def get_extra_context(self):
        return {
            'content_object': self.external_object,
            'hide_link': True,
            'hide_object': True,
            'layer': self.layer,
            'layer_name': self.layer.name,
            'navigation_disable_menus_link_group_object_header': True,
            'navigation_object_list': ('content_object',),
            'no_results_icon': self.layer.icon,
            'no_results_main_link': link_transformation_select.resolve(
                context=RequestContext(
                    dict_={
                        'resolved_object': self.external_object,
                        'layer_name': self.kwargs['layer_name'],
                    }, request=self.request
                )
            ),
            'no_results_text': self.layer.get_empty_results_text(),
            'no_results_title': _(
                message='There are no entries for layer "%(layer_name)s"'
            ) % {'layer_name': self.layer.label},
            'title': _(
                message='Layer "%(layer)s" transformations for: %(object)s'
            ) % {
                'layer': self.layer,
                'object': self.external_object
            }
        }

    def get_source_queryset(self):
        return self.layer.get_transformations_for(obj=self.external_object)


class TransformationSelectView(
    ViewMixinLayer, ExternalContentTypeObjectViewMixin, FormView
):
    form_class = LayerTransformationSelectForm
    template_name = 'appearance/form_container.html'
    view_icon = icon_transformation_select

    def form_valid(self, form):
        transformation_class = BaseTransformation.get(
            name=form.cleaned_data['transformation']
        )
        if transformation_class.arguments:
            return HttpResponseRedirect(
                redirect_to=reverse(
                    kwargs={
                        'app_label': self.kwargs['app_label'],
                        'model_name': self.kwargs['model_name'],
                        'object_id': self.kwargs['object_id'],
                        'layer_name': self.kwargs['layer_name'],
                        'transformation_name': form.cleaned_data[
                            'transformation'
                        ]
                    }, viewname='converter:transformation_create'
                )
            )
        else:
            object_layer, created = ObjectLayer.objects.get_for(
                obj=self.external_object, layer=self.layer
            )
            object_layer.transformations.create(
                name=form.cleaned_data['transformation']
            )

            messages.success(
                message=_(message='Transformation created successfully.'),
                request=self.request
            )

            return HttpResponseRedirect(
                redirect_to=reverse(
                    kwargs={
                        'app_label': self.kwargs['app_label'],
                        'model_name': self.kwargs['model_name'],
                        'object_id': self.kwargs['object_id'],
                        'layer_name': self.kwargs['layer_name']
                    }, viewname='converter:transformation_list'
                )
            )

    def get_external_object_permission(self):
        return self.layer.get_permission(action='select')

    def get_extra_context(self):
        return {
            'content_object': self.external_object,
            'layer': self.layer,
            'layer_name': self.kwargs['layer_name'],
            'navigation_object_list': ('content_object',),
            'submit_label': _(message='Select'),
            'title': _(
                message='Select new layer "%(layer)s" transformation '
                'for: %(object)s'
            ) % {
                'layer': self.layer,
                'object': self.external_object
            }
        }

    def get_form_extra_kwargs(self):
        return {'layer': self.layer}



from django.contrib import messages
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventType
from mayan.apps.events.models import StoredEventType
from mayan.apps.views.generics import FormView
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from ..forms.workflow_template_transition_trigger_forms import (
    WorkflowTransitionTriggerEventRelationshipFormSet
)
from ..icons import icon_workflow_template_transition_triggers
from ..models import WorkflowTransition
from ..permissions import permission_workflow_template_view


class WorkflowTemplateTransitionTriggerEventListView(
    ExternalObjectViewMixin, FormView
):
    external_object_class = WorkflowTransition
    external_object_permission = permission_workflow_template_view
    external_object_pk_url_kwarg = 'workflow_template_transition_id'
    form_class = WorkflowTransitionTriggerEventRelationshipFormSet
    view_icon = icon_workflow_template_transition_triggers

    def form_valid(self, form):
        try:
            for instance in form:
                instance.save()
        except Exception as exception:
            messages.error(
                message=_(
                    message='Error updating workflow transition trigger '
                    'events; %s'
                ) % exception, request=self.request

            )
        else:
            messages.success(
                message=_(
                    message='Workflow transition trigger events updated '
                    'successfully'
                ), request=self.request
            )

        return super().form_valid(form=form)

    def get_extra_context(self):
        return {
            'form_display_mode_table': True,
            'navigation_object_list': ('object', 'workflow'),
            'object': self.external_object,
            'subtitle': _(
                message='Triggers are events that cause this transition to '
                'execute automatically.'
            ),
            'title': _(
                message='Workflow transition trigger events for: %s'
            ) % self.external_object,
            'workflow': self.external_object.workflow
        }

    def get_initial(self):
        obj = self.external_object
        initial = []

        # Return the queryset by name from the sorted list of the class.
        event_type_ids = [
            event_type.id for event_type in EventType.all()
        ]
        queryset_event_types = StoredEventType.objects.filter(
            name__in=event_type_ids
        )

        # Sort queryset in Python by namespace, then by label.
        queryset_event_types = sorted(
            queryset_event_types, key=lambda x: (x.namespace, x.label)
        )

        for event_type in queryset_event_types:
            initial.append(
                {
                    'event_type': event_type,
                    'transition': obj
                }
            )
        return initial

    def get_post_action_redirect(self):
        return reverse(
            kwargs={'workflow_template_id': self.external_object.workflow.pk},
            viewname='document_states:workflow_template_transition_list'
        )



from mayan.apps.document_exports.classes import DocumentVersionExporter
from mayan.apps.document_exports.literals import (
    DOCUMENT_VERSION_EXPORT_MIMETYPE
)
from mayan.apps.storage.utils import NamedTemporaryFile

from .literals import EMAIL_SEPARATORS


def split_recipient_list(recipients, separator_list=None, separator_index=0):
    separator_list = separator_list or EMAIL_SEPARATORS

    try:
        separator = separator_list[separator_index]
    except IndexError:
        return recipients
    else:
        result = []
        for recipient in recipients:
            result.extend(
                recipient.split(separator)
            )

        return split_recipient_list(
            recipients=result, separator_list=separator_list,
            separator_index=separator_index + 1
        )


def get_document_file_content(obj):
    return obj.open()


def get_document_file_mime_type(obj):
    return obj.mimetype


def get_document_version_content(obj):
    class TemporaryExportedDocumentVersion:
        def __init__(self, obj):
            self.obj = obj

        def __enter__(self):
            self.file_object = NamedTemporaryFile(delete=False)
            document_version_exporter = DocumentVersionExporter(
                document_version=obj
            )
            document_version_exporter.export(file_object=self.file_object)
            self.file_object.seek(0)
            return self.file_object

        def __exit__(self, *exc):
            self.file_object.close()

    return TemporaryExportedDocumentVersion(obj=obj)


def get_document_version_mime_type(obj):
    return DOCUMENT_VERSION_EXPORT_MIMETYPE



from django.apps import apps
from django.contrib.auth import get_user_model
from django.db import models
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.app_manager.apps import MayanAppConfig
from mayan.apps.common.menus import (
    menu_list_facet, menu_multi_item, menu_object, menu_return,
    menu_secondary, menu_tools, menu_topbar
)
from mayan.apps.forms import column_widgets
from mayan.apps.navigation.source_columns import SourceColumn

from .classes import EventTypeNamespace
from .html_widgets import widget_event_actor_link, widget_event_type_link
from .links import (
    link_event_list, link_event_list_clear, link_event_list_export,
    link_event_type_subscription_list, link_notification_delete_multiple,
    link_notification_delete_single, link_notification_list,
    link_notification_mark_read, link_notification_mark_read_all,
    link_object_event_list_clear, link_object_event_list_export,
    link_user_object_subscription_list
)


class EventsApp(MayanAppConfig):
    app_namespace = 'events'
    app_url = 'events'
    has_rest_api = True
    has_tests = True
    name = 'mayan.apps.events'
    verbose_name = _(message='Events')

    def ready(self):
        super().ready()

        Action = apps.get_model(app_label='actstream', model_name='Action')
        Notification = self.get_model(model_name='Notification')
        ObjectEventSubscription = self.get_model(
            model_name='ObjectEventSubscription'
        )
        StoredEventType = self.get_model(model_name='StoredEventType')

        User = get_user_model()

        EventTypeNamespace.load_modules()

        Action.add_to_class(
            name='_ordering_fields', value=('timestamp',)
        )

        # Typecast the related field because actstream uses CharFields for
        # the object_id the action_object, actor, and target fields.
        ModelPermission.register_inheritance(
            fk_field_cast=models.CharField, model=Action,
            related='action_object'
        )
        ModelPermission.register_inheritance(
            fk_field_cast=models.CharField, model=Action, related='actor'
        )
        ModelPermission.register_inheritance(
            fk_field_cast=models.CharField, model=Action, related='target'
        )

        ModelPermission.register_inheritance(
            fk_field_cast=models.CharField, model=Notification,
            related='action__action_object'
        )
        ModelPermission.register_inheritance(
            fk_field_cast=models.CharField, model=Notification,
            related='action__actor'
        )
        ModelPermission.register_inheritance(
            fk_field_cast=models.CharField, model=Notification,
            related='action__target'
        )

        ModelPermission.register_inheritance(
            model=ObjectEventSubscription, related='content_object'
        )

        # Add labels to Action model, they are not marked translatable in the
        # upstream package.
        SourceColumn(
            attribute='timestamp', is_identifier=True,
            is_sortable=True, label=_(message='Date and time'), name='timestamp',
            source=Action
        )
        SourceColumn(
            func=widget_event_actor_link, label=_(message='Actor'),
            include_label=True, source=Action
        )
        SourceColumn(
            func=widget_event_type_link, label=_(message='Event'),
            include_label=True, name='event_type', source=Action
        )
        SourceColumn(
            attribute='target', label=_(message='Target'), include_label=True,
            name='target', source=Action, widget=column_widgets.ObjectLinkWidget
        )
        SourceColumn(
            attribute='action_object', label=_(message='Action object'),
            include_label=True, source=Action, widget=column_widgets.ObjectLinkWidget
        )

        # Stored event type

        SourceColumn(
            attribute='namespace', label=_(message='Namespace'),
            source=StoredEventType

        )
        SourceColumn(
            attribute='label', label=_(message='Label'), source=StoredEventType
        )

        # Notification

        SourceColumn(
            attribute='action__timestamp', is_identifier=True,
            is_sortable=True, label=_(message='Date and time'), source=Notification
        )
        SourceColumn(
            func=widget_event_actor_link, include_label=True,
            label=_(message='Actor'), kwargs={'attribute': '"action"'},
            source=Notification
        )
        SourceColumn(
            func=widget_event_type_link, include_label=True,
            label=_(message='Event'), kwargs={'attribute': '"action"'},
            source=Notification
        )
        SourceColumn(
            attribute='action.target', include_label=True,
            label=_(message='Target'), source=Notification,
            widget=column_widgets.ObjectLinkWidget
        )
        SourceColumn(
            attribute='action.action_object', include_label=True,
            label=_(message='Action object'), source=Notification,
            widget=column_widgets.ObjectLinkWidget
        )
        SourceColumn(
            attribute='read', include_label=True, is_sortable=True,
            label=_(message='Seen'), source=Notification,
            widget=column_widgets.TwoStateWidget
        )

        # Object event subscription

        SourceColumn(
            attribute='content_object', include_label=True,
            label=_(message='Object'), source=ObjectEventSubscription,
            widget=column_widgets.ObjectLinkWidget
        )
        SourceColumn(
            attribute='stored_event_type', include_label=True,
            label=_(message='Event type'), source=ObjectEventSubscription
        )

        # Clear

        menu_secondary.bind_links(
            links=(link_event_list_clear,),
            sources=(
                'events:event_list', 'events:event_list_clear'
            )
        )
        menu_secondary.bind_links(
            links=(link_object_event_list_clear,),
            sources=(
                'events:object_event_list', 'events:object_event_list_clear',
                'events:object_event_list_export'
            )
        )

        # Export

        menu_secondary.bind_links(
            links=(link_event_list_export,),
            sources=(
                'events:event_list', 'events:event_list_export'
            )
        )
        menu_secondary.bind_links(
            links=(link_object_event_list_export,),
            sources=(
                'events:object_event_list', 'events:object_event_list_clear',
                'events:object_event_list_export'
            )
        )

        # Notification

        menu_object.bind_links(
            links=(
                link_notification_delete_single, link_notification_mark_read,
            ), sources=(Notification,)
        )
        menu_multi_item.bind_links(
            links=(
                link_notification_delete_multiple,
            ), sources=('events:user_notifications_list',)
        )
        menu_secondary.bind_links(
            links=(link_notification_mark_read_all,),
            sources=(
                'events:notification_mark_read',
                'events:notification_mark_read_all',
                'events:user_notifications_list'
            )
        )

        # Subscription

        menu_list_facet.bind_links(
            links=(
                link_event_type_subscription_list,
                link_user_object_subscription_list,
            ), sources=(User,)
        )

        # Other

        menu_return.bind_links(
            links=(link_event_list,), sources=(
                'events:event_list', 'events:event_list_clear',
                'events:event_list_export'
            )
        )
        menu_topbar.bind_links(
            links=(link_notification_list,), position=30
        )
        menu_tools.bind_links(
            links=(link_event_list,)
        )



from django.apps import apps
from django.db.utils import OperationalError, ProgrammingError

from .literals import STORAGE_NAME_SOURCE_CACHE_FOLDER


def callback_update_source_cache_size(setting):
    Cache = apps.get_model(app_label='file_caching', model_name='Cache')

    try:
        cache = Cache.objects.get(
            defined_storage_name=STORAGE_NAME_SOURCE_CACHE_FOLDER
        )
    except (Cache.DoesNotExist, OperationalError, ProgrammingError):
        """
        Non fatal. Non initialized installation. Ignore exception.
        This will generate an error log entry similar to this.
        2023-12-12 09:12:46.925 UTC [79] ERROR:  relation "file_caching_cache" does not exist at character 121
        2023-12-12 09:12:46.925 UTC [79] STATEMENT:  SELECT "file_caching_cache"."id", "file_caching_cache"."defined_storage_name", "file_caching_cache"."maximum_size" FROM "file_caching_cache" WHERE "file_caching_cache"."defined_storage_name" = 'sources__source_cache' LIMIT 21
        """
    else:
        cache.maximum_size = setting.value
        cache.save()



import logging

from django.contrib import messages
from django.template import RequestContext
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext_lazy as _

from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerMethodAfter
from mayan.apps.storage.views.download_views import ViewSingleObjectDownload
from mayan.apps.views.generics import (
    ConfirmView, SingleObjectCreateView, SingleObjectDeleteView,
    SingleObjectDetailView, SingleObjectListView, SimpleView
)

from .events import event_key_downloaded
from .forms import KeyDetailForm, KeySearchForm
from .icons import (
    icon_key_delete, icon_key_detail, icon_key_download, icon_key_setup,
    icon_key_upload, icon_keyserver_search, icon_private_key_list,
    icon_public_key_list
)
from .links import link_key_query, link_key_upload
from .literals import KEY_TYPE_PUBLIC
from .models import Key
from .permissions import (
    permission_key_delete, permission_key_download, permission_key_receive,
    permission_key_upload, permission_key_view, permission_keyserver_query
)

logger = logging.getLogger(name=__name__)


class KeyDeleteView(SingleObjectDeleteView):
    model = Key
    object_permission = permission_key_delete
    pk_url_kwarg = 'key_id'
    view_icon = icon_key_delete

    def get_extra_context(self):
        return {
            'title': _(message='Delete key: %s') % self.object
        }

    def get_post_action_redirect(self):
        if self.object.key_type == KEY_TYPE_PUBLIC:
            return reverse_lazy(viewname='django_gpg:key_public_list')
        else:
            return reverse_lazy(viewname='django_gpg:key_private_list')


class KeyDetailView(SingleObjectDetailView):
    form_class = KeyDetailForm
    model = Key
    object_permission = permission_key_view
    pk_url_kwarg = 'key_id'
    view_icon = icon_key_detail

    def get_extra_context(self):
        return {
            'title': _(message='Details for key: %s') % self.object
        }


class KeyDownloadView(ViewSingleObjectDownload):
    model = Key
    object_permission = permission_key_download
    pk_url_kwarg = 'key_id'
    view_icon = icon_key_download

    @method_event(
        actor='request.user',
        event_manager_class=EventManagerMethodAfter,
        event=event_key_downloaded,
        target='object'
    )
    def get_download_file_object(self):
        """
        Passthrough code to ensure the download event
        is triggered.
        """
        return super().get_download_file_object()

    def get_download_filename(self):
        return self.object.key_id


class KeyReceive(ConfirmView):
    post_action_redirect = reverse_lazy(
        viewname='django_gpg:key_public_list'
    )
    view_icon = icon_key_upload
    view_permission = permission_key_receive

    def get_extra_context(self):
        return {
            'message': _(message='Import key ID: %s?') % self.kwargs['key_id'],
            'title': _(message='Import key')
        }

    def view_action(self):
        try:
            Key.objects.receive_key(
                key_id=self.kwargs['key_id']
            )
        except Exception as exception:
            messages.error(
                message=_(
                    'Unable to import key: %(key_id)s; %(error)s'
                ) % {
                    'key_id': self.kwargs['key_id'],
                    'error': exception,
                }, request=self.request
            )
        else:
            messages.success(
                message=_(message='Successfully received key: %(key_id)s') % {
                    'key_id': self.kwargs['key_id'],
                }, request=self.request
            )


class KeyQueryResultView(SingleObjectListView):
    view_icon = icon_keyserver_search
    view_permission = permission_keyserver_query

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_key_setup,
            'no_results_main_link': link_key_query.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'Use names, last names, key ids or emails to search '
                'public keys to import from the keyserver.'
            ),
            'no_results_title': _(
                'No results returned'
            ),
            'title': _(message='Key query results')
        }

    def get_source_queryset(self):
        term = self.request.GET.get('term')
        if term:
            return Key.objects.search(query=term)
        else:
            return ()


class KeyQueryView(SimpleView):
    template_name = 'appearance/form_container.html'
    view_icon = icon_keyserver_search
    view_permission = permission_keyserver_query

    def get_extra_context(self):
        return {
            'form': self.get_form(),
            'form_action': reverse(viewname='django_gpg:key_query_results'),
            'submit_method': 'GET',
            'title': _(message='Query key server')
        }

    def get_form(self):
        if ('term' in self.request.GET) and self.request.GET['term'].strip():
            term = self.request.GET['term']
            return KeySearchForm(
                initial={'term': term}
            )
        else:
            return KeySearchForm()


class KeyUploadView(SingleObjectCreateView):
    fields = ('key_data',)
    model = Key
    post_action_redirect = reverse_lazy(
        viewname='django_gpg:key_public_list'
    )
    view_icon = icon_key_upload
    view_permission = permission_key_upload

    def get_extra_context(self):
        return {
            'title': _(message='Upload new key')
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class PrivateKeyListView(SingleObjectListView):
    object_permission = permission_key_view
    source_queryset = Key.objects.private_keys()
    view_icon = icon_private_key_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_private_key_list,
            'no_results_main_link': link_key_upload.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'Private keys are used to signed documents. '
                'Private keys can only be uploaded by the user. '
                'The view to upload private and public keys is the same.'
            ),
            'no_results_title': _(
                'There no private keys'
            ),
            'title': _(message='Private keys')
        }


class PublicKeyListView(SingleObjectListView):
    object_permission = permission_key_view
    source_queryset = Key.objects.public_keys()
    view_icon = icon_public_key_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_public_key_list,
            'no_results_main_link': link_key_upload.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'Public keys are used to verify signed documents. '
                'Public keys can be uploaded by the user or downloaded '
                'from keyservers. The view to upload private and public '
                'keys is the same.'
            ),
            'no_results_title': _(
                'There no public keys'
            ),
            'title': _(message='Public keys')
        }



from django.core.exceptions import NON_FIELD_ERRORS, ValidationError
from django.db import connection, models, transaction
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from mptt.fields import TreeForeignKey
from mptt.models import MPTTModel

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.documents.models.document_models import Document
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import (
    EventManagerMethodAfter, EventManagerSave
)

from .events import (
    event_cabinet_created, event_cabinet_deleted,
    event_cabinet_document_added, event_cabinet_document_removed,
    event_cabinet_edited
)
from .model_mixins import CabinetBusinessLogicMixin


class Cabinet(CabinetBusinessLogicMixin, ExtraDataModelMixin, MPTTModel):
    """
    Model to store a hierarchical tree of document containers. Each container
    can store an unlimited number of documents using an M2M field. Only
    the top level container is can have an ACL. All child container's
    access is delegated to their corresponding root container.
    """
    _ordering_fields = ('label',)

    parent = TreeForeignKey(
        blank=True, db_index=True, null=True, on_delete=models.CASCADE,
        related_name='children', to='self', verbose_name=_(message='Parent')
    )
    label = models.CharField(
        help_text=_(message='A short text used to identify the cabinet.'),
        max_length=128, verbose_name=_(message='Label')
    )
    documents = models.ManyToManyField(
        blank=True, related_name='cabinets', to=Document,
        verbose_name=_(message='Documents')
    )

    class MPTTMeta:
        order_insertion_by = ('label',)

    class Meta:
        # unique_together doesn't work if there is a FK
        # https://code.djangoproject.com/ticket/1751
        unique_together = ('parent', 'label')
        verbose_name = _(message='Cabinet')
        verbose_name_plural = _(message='Cabinets')

    def __str__(self):
        return self.get_full_path()

    @method_event(
        action_object='self',
        event=event_cabinet_document_added,
        event_manager_class=EventManagerMethodAfter
    )
    def _document_add(self, document, user=None):
        self._event_actor = user
        self._event_target = document
        self.documents.add(document)

    @method_event(
        action_object='self',
        event=event_cabinet_document_removed,
        event_manager_class=EventManagerMethodAfter
    )
    def _document_remove(self, document, user=None):
        self._event_actor = user
        self._event_target = document
        self.documents.remove(document)

    @method_event(
        action_object='parent',
        event=event_cabinet_deleted,
        event_manager_class=EventManagerMethodAfter
    )
    def delete(self, *args, **kwargs):
        self._event_actor = getattr(self, '_event_actor', 'parent')

        result = super().delete(*args, **kwargs)

        if not self.parent:
            self._event_ignore = True

        return result

    def get_absolute_url(self):
        return reverse(
            kwargs={'cabinet_id': self.pk},
            viewname='cabinets:cabinet_view'
        )

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'action_object': 'parent',
            'event': event_cabinet_created,
            'target': 'self'
        },
        edited={
            'event': event_cabinet_edited,
            'target': 'self'
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)

    def validate_unique(self, exclude=None):
        """
        Explicit validation of uniqueness of parent+label as the provided
        unique_together check in Meta is not working for all 100% cases
        when there is a FK in the unique_together tuple
        https://code.djangoproject.com/ticket/1751
        """
        with transaction.atomic():
            if connection.vendor == 'oracle':
                queryset = Cabinet.objects.filter(
                    parent=self.parent, label=self.label
                )
            else:
                queryset = Cabinet.objects.select_for_update().filter(
                    parent=self.parent, label=self.label
                )

            if queryset.exists():
                params = {
                    'field_labels': _(message='Parent and Label'),
                    'model_name': _(message='Cabinet')
                }
                raise ValidationError(
                    message={
                        NON_FIELD_ERRORS: [
                            ValidationError(
                                message=_(
                                    message='%(model_name)s with this '
                                    '%(field_labels)s already exists.'
                                ), code='unique_together', params=params
                            )
                        ]
                    }
                )


class CabinetSearchResult(Cabinet):
    """
    Represent a cabinet's search result. This model is a proxy model from
    Cabinet and is used as an alias to map columns to it without having to
    map them to the base Cabinet model.
    """
    class Meta:
        proxy = True
        verbose_name = _(message='Cabinet')
        verbose_name_plural = _(message='Cabinets')


class DocumentCabinet(Cabinet):
    """
    Represent a document's cabinet. This Model is a proxy model from Cabinet
    and is used as an alias to map columns to it without having to map them
    to the base Cabinet model.
    """
    class Meta:
        proxy = True
        verbose_name = _(message='Document cabinet')
        verbose_name_plural = _(message='Document cabinets')



from django.template import TemplateSyntaxError

from rest_framework.exceptions import ValidationError
from rest_framework.permissions import IsAuthenticated
from rest_framework.settings import api_settings

from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import ContentTypeAPIViewMixin

from .classes import AJAXTemplate, ModelTemplating
from .permissions import permission_template_sandbox
from .serializers import (
    AJAXTemplateSerializer, ObjectTemplateSandboxSerializer
)


class APITemplateDetailView(generics.RetrieveAPIView):
    """
    Returns the selected partial template details.
    get: Retrieve the details of the partial template.
    """
    permission_classes = (IsAuthenticated,)
    serializer_class = AJAXTemplateSerializer

    def get_object(self):
        return AJAXTemplate.get(
            name=self.kwargs['name']
        ).render(
            request=self.request
        )


class APITemplateListView(generics.ListAPIView):
    """
    Returns a list of all the available templates.
    """
    permission_classes = (IsAuthenticated,)
    serializer_class = AJAXTemplateSerializer

    def get_source_queryset(self):
        return AJAXTemplate.all(
            rendered=True, request=self.request
        )


class APIObjectTemplateSandboxActionView(
    ContentTypeAPIViewMixin, generics.ObjectActionAPIView
):
    """
    post: Interactive inspection of object properties.
    """
    mayan_object_permission_map = {
        'GET': permission_template_sandbox,
        'POST': permission_template_sandbox
    }
    lookup_url_kwarg = 'object_id'
    serializer_class = ObjectTemplateSandboxSerializer

    def object_action(self, obj, request, serializer):
        template_string = serializer.validated_data['template']

        try:
            result = ModelTemplating.do_render(
                obj=obj, template_string=template_string
            )
        except KeyError as exception:
            raise ValidationError(
                {
                    api_settings.NON_FIELD_ERRORS_KEY: [
                        str(exception)
                    ]
                }, code='invalid'
            )
        except TemplateSyntaxError as exception:
            raise ValidationError(
                {
                    'template': [
                        str(exception)
                    ]
                }, code='invalid'
            )
        else:
            return {'result': result}

    def get_source_queryset(self):
        content_type = self.get_content_type()
        return content_type.get_all_objects_for_this_type()



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('document_parsing', '0006_rename_documentversionparseerror_model')
    ]

    operations = [
        migrations.AlterModelOptions(
            name='documentfileparseerror',
            options={
                'ordering': ('datetime_submitted',),
                'verbose_name': 'Document file parse error',
                'verbose_name_plural': 'Document file parse errors'
            }
        ),
        migrations.AlterField(
            model_name='documentfileparseerror',
            name='document_file',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='parsing_errors', to='documents.DocumentFile',
                verbose_name='Document file'
            )
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster
from mayan.apps.task_manager.classes import Worker
from mayan.settings.literals import (  # NOQA
    DOCKER_USER_GID, DOCKER_USER_UID, MAYAN_WORKER_A_CONCURRENCY,
    MAYAN_WORKER_A_MAX_MEMORY_PER_CHILD, MAYAN_WORKER_A_MAX_TASKS_PER_CHILD,
    MAYAN_WORKER_B_CONCURRENCY, MAYAN_WORKER_B_MAX_MEMORY_PER_CHILD,
    MAYAN_WORKER_B_MAX_TASKS_PER_CHILD, MAYAN_WORKER_C_CONCURRENCY,
    MAYAN_WORKER_C_MAX_MEMORY_PER_CHILD, MAYAN_WORKER_C_MAX_TASKS_PER_CHILD,
    MAYAN_WORKER_D_CONCURRENCY, MAYAN_WORKER_D_MAX_MEMORY_PER_CHILD,
    MAYAN_WORKER_D_MAX_TASKS_PER_CHILD, MAYAN_WORKER_E_CONCURRENCY,
    MAYAN_WORKER_E_MAX_MEMORY_PER_CHILD, MAYAN_WORKER_E_MAX_TASKS_PER_CHILD
)

from .literals import (
    DEFAULT_APT_INSTALLS, DEFAULT_DOCKER_SCRIPT_POST_SETUP,
    DEFAULT_DOCKER_SCRIPT_PRE_SETUP, DEFAULT_DOCKER_WAIT,
    DEFAULT_PIP_INSTALLS, DEFAULT_SKIP_CHOWN_ON_STARTUP,
    worker_setting_map
)

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Docker'), name='docker'
)

setting_namespace.do_setting_add(
    default=DEFAULT_APT_INSTALLS, global_name='MAYAN_APT_INSTALLS',
    help_text=_(
        message='Specifies a list of .deb packages to be installed via '
        'APT when the container is first created. The installed packages '
        'are not lost when the image is stopped.'
    )
)
setting_pip_install = setting_namespace.do_setting_add(
    default=DEFAULT_PIP_INSTALLS, global_name='MAYAN_PIP_INSTALLS',
    help_text=_(
        message='Specifies a list of Python packages to be installed via '
        'pip. Packages will be downloaded from the Python Package Index '
        '(https://pypi.python.org) by default.'
    )
)
setting_script_post_setup = setting_namespace.do_setting_add(
    default=DEFAULT_DOCKER_SCRIPT_POST_SETUP,
    global_name='MAYAN_DOCKER_SCRIPT_POST_SETUP', help_text=_(
        message='Executed after the container\'s environment variables '
        'are configured, after the UID/GID setup, extra OS package '
        'installations and extra Python library installations but before '
        'launching the Mayan EDMS stack. Executes the content as a script '
        'or call a script using the value as the filename.'
    )
)
setting_script_pre_setup = setting_namespace.do_setting_add(
    default=DEFAULT_DOCKER_SCRIPT_PRE_SETUP,
    global_name='MAYAN_DOCKER_SCRIPT_PRE_SETUP', help_text=_(
        message='Executed after the container\'s environment variables are '
        'configured but before the UID/GID setup, extra OS package '
        'installations and extra Python library installations. Executes '
        'the content as a script or call a script using the value as the '
        'filename.'
    )
)
setting_namespace.do_setting_add(
    default=DEFAULT_SKIP_CHOWN_ON_STARTUP,
    global_name='MAYAN_SKIP_CHOWN_ON_STARTUP', help_text=_(
        message='Setting this environment variable to true, will make the '
        'entrypoint script skip the initial `chown` command on the media '
        'folder at `/var/lib/mayan`.'
    )
)
setting_namespace.do_setting_add(
    default=DOCKER_USER_GID,
    global_name='MAYAN_USER_GID', help_text=_(
        message='Changes the GID of the `mayan` user internal to the Docker '
        'container.'
    )
)
setting_namespace.do_setting_add(
    default=DOCKER_USER_UID,
    global_name='MAYAN_USER_UID', help_text=_(
        message='Changes the UID of the `mayan` user internal to the Docker '
        'container.'
    )
)

setting_namespace.do_setting_add(
    default=DEFAULT_DOCKER_WAIT,
    global_name='MAYAN_DOCKER_WAIT', help_text=_(
        message='Make the Docker container wait for a host and port to '
        'become available. Multiple hosts and port combinations are '
        'supported.'
    )
)

for worker in Worker.all():
    worker_name = worker.name.upper()

    for setting_name, settings_attributes in worker_setting_map:
        global_name = 'MAYAN_{}_{}'.format(
            worker_name, setting_name.upper()
        )

        default_name = 'MAYAN_{}_{}'.format(
            worker_name, settings_attributes['default_name'].upper()
        )

        default = globals()[default_name]

        help_text = settings_attributes['help_text']

        setting_namespace.do_setting_add(
            default=default, global_name=global_name, help_text=help_text
        )



from django.template import Library

from ..models import Announcement

register = Library()


@register.inclusion_tag(
    filename='announcements/announcements.html', name='announcements'
)
def tag_announcements():
    return {'announcements': Announcement.objects.get_for_now()}



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('document_indexing', '0014_auto_20180823_2353')
    ]

    operations = [
        migrations.AlterField(
            model_name='index',
            name='document_types',
            field=models.ManyToManyField(
                related_name='indexes', to='documents.DocumentType',
                verbose_name='Document types'
            ),
        )
    ]



from django.template import RequestContext
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from mayan.apps.views.generics import (
    SingleObjectCreateView, SingleObjectDeleteView, SingleObjectEditView,
    SingleObjectListView
)
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from ..forms.workflow_template_transition_field_forms import (
    WorkflowTransitionFieldForm
)
from ..icons import (
    icon_workflow_template_transition_field,
    icon_workflow_template_transition_field_create,
    icon_workflow_template_transition_field_delete,
    icon_workflow_template_transition_field_edit,
    icon_workflow_template_transition_field_list
)
from ..links import link_workflow_template_transition_field_create
from ..models import WorkflowTransition, WorkflowTransitionField
from ..permissions import (
    permission_workflow_template_edit, permission_workflow_template_view
)


class WorkflowTemplateTransitionFieldCreateView(
    ExternalObjectViewMixin, SingleObjectCreateView
):
    external_object_class = WorkflowTransition
    external_object_permission = permission_workflow_template_edit
    external_object_pk_url_kwarg = 'workflow_template_transition_id'
    form_class = WorkflowTransitionFieldForm
    view_icon = icon_workflow_template_transition_field_create

    def get_extra_context(self):
        return {
            'navigation_object_list': ('transition', 'workflow'),
            'transition': self.external_object,
            'title': _(
                message='Create a field for workflow transition: %s'
            ) % self.external_object,
            'workflow': self.external_object.workflow
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'transition': self.external_object
        }

    def get_queryset(self):
        return self.external_object.fields.all()

    def get_post_action_redirect(self):
        return reverse(
            kwargs={
                'workflow_template_transition_id': self.external_object.pk
            },
            viewname='document_states:workflow_template_transition_field_list'
        )


class WorkflowTemplateTransitionFieldDeleteView(SingleObjectDeleteView):
    model = WorkflowTransitionField
    object_permission = permission_workflow_template_edit
    pk_url_kwarg = 'workflow_template_transition_field_id'
    view_icon = icon_workflow_template_transition_field_delete

    def get_extra_context(self):
        return {
            'navigation_object_list': (
                'object', 'workflow_template_transition', 'workflow'
            ),
            'object': self.object,
            'title': _(
                message='Delete workflow transition field: %s'
            ) % self.object,
            'workflow': self.object.transition.workflow,
            'workflow_template_transition': self.object.transition
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_post_action_redirect(self):
        return reverse(
            kwargs={
                'workflow_template_transition_id': self.object.transition.pk
            },
            viewname='document_states:workflow_template_transition_field_list'
        )


class WorkflowTemplateTransitionFieldEditView(SingleObjectEditView):
    form_class = WorkflowTransitionFieldForm
    model = WorkflowTransitionField
    object_permission = permission_workflow_template_edit
    pk_url_kwarg = 'workflow_template_transition_field_id'
    view_icon = icon_workflow_template_transition_field_edit

    def get_extra_context(self):
        return {
            'navigation_object_list': (
                'object', 'workflow_template_transition', 'workflow'
            ),
            'object': self.object,
            'title': _(
                message='Edit workflow transition field: %s'
            ) % self.object,
            'workflow': self.object.transition.workflow,
            'workflow_template_transition': self.object.transition
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_post_action_redirect(self):
        return reverse(
            kwargs={
                'workflow_template_transition_id': self.object.transition.pk
            },
            viewname='document_states:workflow_template_transition_field_list'
        )


class WorkflowTemplateTransitionFieldListView(
    ExternalObjectViewMixin, SingleObjectListView
):
    external_object_class = WorkflowTransition
    external_object_permission = permission_workflow_template_view
    external_object_pk_url_kwarg = 'workflow_template_transition_id'
    view_icon = icon_workflow_template_transition_field_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'navigation_object_list': ('object', 'workflow'),
            'no_results_icon': icon_workflow_template_transition_field,
            'no_results_main_link': link_workflow_template_transition_field_create.resolve(
                context=RequestContext(
                    dict_={
                        'object': self.external_object
                    }, request=self.request
                )
            ),
            'no_results_text': _(
                message='Workflow transition fields allow adding data to the '
                'workflow\'s context. This additional context data can then '
                'be used by other elements of the workflow system like the '
                'workflow state actions.'
            ),
            'no_results_title': _(
                message='There are no fields for this workflow transition'
            ),
            'object': self.external_object,
            'title': _(
                message='Fields for workflow transition: %s'
            ) % self.external_object,
            'workflow': self.external_object.workflow
        }

    def get_source_queryset(self):
        return self.external_object.fields.all()



from django.utils.translation import gettext_lazy as _

from mayan.apps.views.generics import SingleObjectListView

from .apps import MayanAppConfig
from .icons import icon_app_list


class AppListView(SingleObjectListView):
    view_icon = icon_app_list

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'subtitle': _(message='Detected and active Mayan EDMS apps.'),
            'title': _(message='System apps')
        }

    def get_source_queryset(self):
        return MayanAppConfig.get_all()



from rest_framework.generics import get_object_or_404

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.api_views.api_view_mixins import (
    ParentObjectDocumentFileAPIViewMixin
)


class ParentObjectDocumentFileDriverAPIViewMixin(
    ParentObjectDocumentFileAPIViewMixin
):
    def get_document_file_metadata_driver(self, permission=None):
        queryset = self.get_document_file_metadata_driver_queryset()

        if permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=queryset,
                user=self.request.user
            )

        return get_object_or_404(
            queryset=queryset, pk=self.kwargs['driver_id']
        )

    def get_document_file_metadata_driver_queryset(self):
        document_file = self.get_document_file()
        return document_file.file_metadata_drivers.all()



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .literals import DEFAULT_TEMPLATING_TAGS_DANGEROUS_ALLOW_LIST

CHOICES_TEMPLATING_WIDGET_HIGHLIGHT_THEME = (
    'a11y-dark', 'a11y-light', 'agate', 'an-old-hope', 'androidstudio',
    'arduino-light', 'arta', 'ascetic', 'atom-one-dark',
    'atom-one-dark-reasonable', 'atom-one-light', 'base16/3024',
    'base16/apathy', 'base16/apprentice', 'base16/ashes',
    'base16/atelier-cave', 'base16/atelier-cave-light',
    'base16/atelier-dune', 'base16/atelier-dune-light',
    'base16/atelier-estuary', 'base16/atelier-estuary-light',
    'base16/atelier-forest', 'base16/atelier-forest-light',
    'base16/atelier-heath', 'base16/atelier-heath-light',
    'base16/atelier-lakeside', 'base16/atelier-lakeside-light',
    'base16/atelier-plateau', 'base16/atelier-plateau-light',
    'base16/atelier-savanna', 'base16/atelier-savanna-light',
    'base16/atelier-seaside', 'base16/atelier-seaside-light',
    'base16/atelier-sulphurpool', 'base16/atelier-sulphurpool-light',
    'base16/atlas', 'base16/bespin', 'base16/black-metal',
    'base16/black-metal-bathory', 'base16/black-metal-burzum',
    'base16/black-metal-dark-funeral', 'base16/black-metal-gorgoroth',
    'base16/black-metal-immortal', 'base16/black-metal-khold',
    'base16/black-metal-marduk', 'base16/black-metal-mayhem',
    'base16/black-metal-nile', 'base16/black-metal-venom', 'base16/brewer',
    'base16/bright', 'base16/brogrammer', 'base16/brush-trees',
    'base16/brush-trees-dark', 'base16/chalk', 'base16/circus',
    'base16/classic-dark', 'base16/classic-light', 'base16/codeschool',
    'base16/colors', 'base16/cupcake', 'base16/cupertino', 'base16/danqing',
    'base16/darcula', 'base16/dark-violet', 'base16/darkmoss',
    'base16/darktooth', 'base16/decaf', 'base16/default-dark',
    'base16/default-light', 'base16/dirtysea', 'base16/dracula',
    'base16/edge-dark', 'base16/edge-light', 'base16/eighties',
    'base16/embers', 'base16/equilibrium-dark',
    'base16/equilibrium-gray-dark', 'base16/equilibrium-gray-light',
    'base16/equilibrium-light', 'base16/espresso', 'base16/eva',
    'base16/eva-dim', 'base16/flat', 'base16/framer', 'base16/fruit-soda',
    'base16/gigavolt', 'base16/github', 'base16/google-dark',
    'base16/google-light', 'base16/grayscale-dark', 'base16/grayscale-light',
    'base16/green-screen', 'base16/gruvbox-dark-hard',
    'base16/gruvbox-dark-medium', 'base16/gruvbox-dark-pale',
    'base16/gruvbox-dark-soft', 'base16/gruvbox-light-hard',
    'base16/gruvbox-light-medium', 'base16/gruvbox-light-soft',
    'base16/hardcore', 'base16/harmonic16-dark', 'base16/harmonic16-light',
    'base16/heetch-dark', 'base16/heetch-light', 'base16/helios',
    'base16/hopscotch', 'base16/horizon-dark', 'base16/horizon-light',
    'base16/humanoid-dark', 'base16/humanoid-light', 'base16/ia-dark',
    'base16/ia-light', 'base16/icy-dark', 'base16/ir-black',
    'base16/isotope', 'base16/kimber', 'base16/london-tube',
    'base16/macintosh', 'base16/marrakesh', 'base16/materia',
    'base16/material', 'base16/material-darker', 'base16/material-lighter',
    'base16/material-palenight', 'base16/material-vivid',
    'base16/mellow-purple', 'base16/mexico-light', 'base16/mocha',
    'base16/monokai', 'base16/nebula', 'base16/nord', 'base16/nova',
    'base16/ocean', 'base16/oceanicnext', 'base16/one-light',
    'base16/onedark', 'base16/outrun-dark', 'base16/papercolor-dark',
    'base16/papercolor-light', 'base16/paraiso', 'base16/pasque',
    'base16/phd', 'base16/pico', 'base16/pop', 'base16/porple',
    'base16/qualia', 'base16/railscasts', 'base16/rebecca',
    'base16/ros-pine', 'base16/ros-pine-dawn', 'base16/ros-pine-moon',
    'base16/sagelight', 'base16/sandcastle', 'base16/seti-ui',
    'base16/shapeshifter', 'base16/silk-dark', 'base16/silk-light',
    'base16/snazzy', 'base16/solar-flare', 'base16/solar-flare-light',
    'base16/solarized-dark', 'base16/solarized-light', 'base16/spacemacs',
    'base16/summercamp', 'base16/summerfruit-dark',
    'base16/summerfruit-light', 'base16/synth-midnight-terminal-dark',
    'base16/synth-midnight-terminal-light', 'base16/tango', 'base16/tender',
    'base16/tomorrow', 'base16/tomorrow-night', 'base16/twilight',
    'base16/unikitty-dark', 'base16/unikitty-light', 'base16/vulcan',
    'base16/windows-10', 'base16/windows-10-light', 'base16/windows-95',
    'base16/windows-95-light', 'base16/windows-high-contrast',
    'base16/windows-high-contrast-light', 'base16/windows-nt',
    'base16/windows-nt-light', 'base16/woodland', 'base16/xcode-dusk',
    'base16/zenburn', 'brown-paper', 'codepen-embed', 'color-brewer',
    'dark', 'default', 'devibeans', 'docco', 'far', 'felipec',
    'foundation', 'github', 'github-dark', 'github-dark-dimmed', 'gml',
    'googlecode', 'gradient-dark', 'gradient-light', 'grayscale', 'hybrid',
    'idea', 'intellij-light', 'ir-black', 'isbl-editor-dark',
    'isbl-editor-light', 'kimbie-dark', 'kimbie-light', 'lightfair',
    'lioshi', 'magula', 'mono-blue', 'monokai', 'monokai-sublime',
    'night-owl', 'nnfx-dark', 'nnfx-light', 'nord', 'obsidian',
    'panda-syntax-dark', 'panda-syntax-light', 'paraiso-dark',
    'paraiso-light', 'pojoaque', 'purebasic', 'qtcreator-dark',
    'qtcreator-light', 'rainbow', 'routeros', 'school-book',
    'shades-of-purple', 'srcery', 'stackoverflow-dark',
    'stackoverflow-light', 'sunburst', 'tokyo-night-dark',
    'tokyo-night-light', 'tomorrow-night-blue', 'tomorrow-night-bright',
    'vs', 'vs2015', 'xcode', 'xt256'
)

DEFAULT_TEMPLATING_WIDGET_HIGHLIGHT_THEME = 'vs2015'

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Templating'), name='templating'
)

setting_templating_dangerous_tags_allow_list = setting_namespace.do_setting_add(
    default=DEFAULT_TEMPLATING_TAGS_DANGEROUS_ALLOW_LIST,
    global_name='TEMPLATING_TAGS_DANGEROUS_ALLOW_LIST', help_text=_(
        message='A comma separated list of dangerous templating tags and '
        'filters that are allowed.'
    )
)
setting_templating_widget_highlight_theme = setting_namespace.do_setting_add(
    choices=CHOICES_TEMPLATING_WIDGET_HIGHLIGHT_THEME,
    default=DEFAULT_TEMPLATING_WIDGET_HIGHLIGHT_THEME,
    global_name='TEMPLATING_WIDGET_HIGHLIGHT_THEME', help_text=_(
        message='Theme name of the CSS stylesheet used to change the '
        'visual appearance of the template preview.'
    )
)



from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import (
    APIViewMixinOwnerPlusFilteredQueryset
)

from ..events import event_download_file_downloaded
from ..models import DownloadFile
from ..permissions import (
    permission_download_file_delete, permission_download_file_download,
    permission_download_file_view
)
from ..serializers import DownloadFileSerializer

from .base import APIObjectDownloadView


class APIDownloadFileDetailView(
    APIViewMixinOwnerPlusFilteredQueryset, generics.RetrieveDestroyAPIView
):
    """
    delete: Delete the selected download file.
    get: Return the details of the selected download file.
    """
    lookup_url_kwarg = 'download_file_id'
    mayan_object_permission_map = {
        'DELETE': permission_download_file_delete,
        'GET': permission_download_file_view
    }
    serializer_class = DownloadFileSerializer
    source_queryset = DownloadFile.objects.all()

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class APIDownloadFileDownloadView(
    APIViewMixinOwnerPlusFilteredQueryset, APIObjectDownloadView
):
    """
    get: Download a download file.
    """
    download_event_type = event_download_file_downloaded
    lookup_url_kwarg = 'download_file_id'
    mayan_object_permission_map = {'GET': permission_download_file_download}
    model = DownloadFile

    def get_download_event_target(self):
        return self.get_object()

    def get_download_file_object(self):
        instance = self.get_object()
        return instance.open(mode='rb')

    def get_download_filename(self):
        return self.get_object().filename

    def get_serializer(self, *args, **kwargs):
        return None

    def get_serializer_class(self):
        return None

    def retrieve(self, request, *args, **kwargs):
        return self.render_to_response()


class APIDownloadFileListView(
    APIViewMixinOwnerPlusFilteredQueryset, generics.ListAPIView
):
    """
    get: Returns a list of all the download files.
    """
    mayan_object_permission_map = {'GET': permission_download_file_view}
    serializer_class = DownloadFileSerializer
    source_queryset = DownloadFile.objects.all()

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}



import logging

from django.template import RequestContext
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from mayan.apps.views.generics import (
    MultipleObjectConfirmActionView, SingleObjectCreateView,
    SingleObjectEditView, SingleObjectListView
)

from .icons import (
    icon_announcement_create, icon_announcement_delete_multiple,
    icon_announcement_edit, icon_announcement_list
)
from .links import link_announcement_create
from .models import Announcement
from .permissions import (
    permission_announcement_create, permission_announcement_delete,
    permission_announcement_edit, permission_announcement_view
)

logger = logging.getLogger(name=__name__)


class AnnouncementCreateView(SingleObjectCreateView):
    fields = ('label', 'text', 'enabled', 'start_datetime', 'end_datetime')
    model = Announcement
    view_icon = icon_announcement_create
    view_permission = permission_announcement_create

    def get_extra_context(self):
        return {
            'title': _(message='Create announcement')
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class AnnouncementDeleteView(MultipleObjectConfirmActionView):
    error_message = _(
        message='Error deleting announcement "%(instance)s"; %(exception)s'
    )
    model = Announcement
    object_permission = permission_announcement_delete
    pk_url_kwarg = 'announcement_id'
    post_action_redirect = reverse_lazy(
        viewname='announcements:announcement_list'
    )
    success_message_plural = _(
        message='%(count)d announcements deleted successfully.'
    )
    success_message_single = _(
        message='Announcement "%(object)s" deleted successfully.'
    )
    success_message_singular = _(
        message='%(count)d announcement deleted successfully.'
    )
    title_plural = _(message='Delete the %(count)d selected announcements.')
    title_single = _(message='Delete announcement: %(object)s.')
    title_singular = _(message='Delete the %(count)d selected announcement.')
    view_icon = icon_announcement_delete_multiple

    def get_extra_context(self):
        context = {'delete_view': True}

        if self.object_list.count() == 1:
            context.update(
                {
                    'object': self.object_list.first()
                }
            )

        return context

    def object_action(self, instance, form=None):
        instance.delete()


class AnnouncementEditView(SingleObjectEditView):
    fields = ('label', 'text', 'enabled', 'start_datetime', 'end_datetime')
    model = Announcement
    object_permission = permission_announcement_edit
    pk_url_kwarg = 'announcement_id'
    post_action_redirect = reverse_lazy(
        viewname='announcements:announcement_list'
    )
    view_icon = icon_announcement_edit

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _(message='Edit announcement: %s') % self.object
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class AnnouncementListView(SingleObjectListView):
    model = Announcement
    object_permission = permission_announcement_view
    view_icon = icon_announcement_list

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': icon_announcement_list,
            'no_results_main_link': link_announcement_create.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                message='Announcements are displayed in the login view. You can use '
                'announcements to convey information about your '
                'organization, announcements or usage guidelines for '
                'your users.'
            ),
            'no_results_title': _(message='No announcements available'),
            'title': _(message='Announcements')
        }



from django.utils.translation import gettext_lazy as _

from mayan.apps.app_manager.apps import MayanAppConfig


class IconsApp(MayanAppConfig):
    app_namespace = 'icons'
    app_url = 'icons'
    has_static_media = True
    name = 'mayan.apps.icons'
    static_media_ignore_patterns = (
        'icons/node_modules/@fortawesome/fontawesome-free/less/*',
        'icons/node_modules/@fortawesome/fontawesome-free/metadata/*',
        'icons/node_modules/@fortawesome/fontawesome-free/sprites/*',
        'icons/node_modules/@fortawesome/fontawesome-free/svgs/*',
    )
    verbose_name = _(message='Icons')



from django.utils.translation import gettext_lazy as _

from mayan.apps.app_manager.apps import MayanAppConfig


class NavigationApp(MayanAppConfig):
    has_tests = True
    name = 'mayan.apps.navigation'
    verbose_name = _(message='Navigation')



import logging

from django.conf import settings
from django.contrib import messages
from django.http import HttpResponseRedirect, JsonResponse
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _

from mayan.apps.views.generics import MultiFormView
from mayan.apps.views.utils import request_is_ajax
from mayan.apps.views.view_mixins import ViewIconMixin

from .view_mixins import SourceActionViewMixin, SourceLinkNavigationViewMixin

logger = logging.getLogger(name=__name__)


class UploadBaseView(
    SourceActionViewMixin, SourceLinkNavigationViewMixin, ViewIconMixin,
    MultiFormView
):
    prefixes = {'document_form': 'document', 'source_form': 'source'}
    template_name = 'appearance/form_container.html'

    def dispatch(self, request, *args, **kwargs):
        self.queryset_source_valid = self.get_queryset_source_valid()

        if not self.queryset_source_valid.exists():
            messages.error(
                message=_(
                    message='There are no enabled sources that support this '
                    'operation. Create a new one or enable and existing one.'
                ), request=request
            )
            return HttpResponseRedirect(
                redirect_to=reverse(viewname='sources:source_list')
            )

        self.source = self.get_source()

        try:
            return super().dispatch(request=request, *args, **kwargs)
        except Exception as exception:
            if settings.DEBUG or settings.TESTING:
                raise
            elif request_is_ajax(request=request):
                return JsonResponse(
                    data={
                        'error': force_str(s=exception)
                    }, status=500
                )
            else:
                raise

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        backend_instance = self.source.get_backend_instance()
        context['source'] = self.source

        context.update(
            backend_instance.get_view_context(
                context=context, request=self.request
            )
        )

        return context

    def get_form_classes(self):
        result = {
            'document_form': self.document_form
        }

        backend_instance = self.source.get_backend_instance()

        view_source_action = self.get_view_source_action()

        action = self.source.get_action(name=view_source_action)

        source_form = backend_instance.get_upload_form_class(action=action)

        if source_form:
            result['source_form'] = source_form

        return result

    def get_source(self):
        queryset = self.queryset_source_valid

        if 'source_id' in self.kwargs:
            pk = self.kwargs['source_id']
        else:
            first_source = queryset.first()
            if first_source:
                pk = queryset.first().pk
            else:
                pk = None

        return get_object_or_404(klass=queryset, pk=pk)

    def get_source_link_action(self):
        return self.get_view_source_action()

    def get_source_link_permission(self):
        return self.object_permission

    def get_source_link_queryset(self):
        return self.queryset_source_valid

    def get_view_source_action(self):
        return self.view_source_action



from django.core.exceptions import ValidationError
from django.utils.text import format_lazy

from mayan.apps.common.class_mixins import AppsModuleLoaderMixin
from mayan.apps.forms.literals import EMPTY_LABEL


class MetadataTypeParserMetaclass(type):
    _registry = {}

    def __new__(mcs, name, bases, attrs):
        new_class = super().__new__(
            mcs, name, bases, attrs
        )
        if new_class.__module__ != 'mayan.apps.metadata.classes':
            mcs._registry[
                '{}.{}'.format(new_class.__module__, name)
            ] = new_class

        return new_class


class MetadataTypeValidatorMetaclass(type):
    _registry = {}

    def __new__(mcs, name, bases, attrs):
        new_class = super().__new__(
            mcs, name, bases, attrs
        )
        if new_class.__module__ != 'mayan.apps.metadata.classes':
            mcs._registry[
                '{}.{}'.format(new_class.__module__, name)
            ] = new_class

        return new_class


class MetadataTypeModuleMixin(AppsModuleLoaderMixin):
    arguments = ()

    @classmethod
    def get_all(cls):
        return cls._registry.values()

    @classmethod
    def get_choices(cls, add_blank=False):
        choices = [
            (
                entry.get_import_path(), entry.get_full_label()
            ) for entry in cls.get_all()
        ]
        choices.sort(
            key=lambda x: x[1]
        )

        if add_blank:
            choices.insert(
                0, (None, EMPTY_LABEL)
            )

        return choices

    @classmethod
    def get_import_path(cls):
        return cls.__module__ + '.' + cls.__name__

    @classmethod
    def get_full_label(cls):
        arguments_string = ', '.join(cls.arguments)
        if arguments_string:
            arguments_template = '(arguments: {})'.format(arguments_string)
        else:
            arguments_template = ''

        return format_lazy(
            '{label} {arguments_template}'.format(
                arguments_template=arguments_template, label=cls.label
            )
        )

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def execute(self, input_data):
        raise NotImplementedError


class MetadataParser(
    MetadataTypeModuleMixin, metaclass=MetadataTypeParserMetaclass
):
    _loader_module_name = 'metadata_parsers'

    def parse(self, input_data):
        try:
            return self.execute(input_data)
        except Exception as exception:
            raise ValidationError(message=exception)


class MetadataValidator(
    MetadataTypeModuleMixin, metaclass=MetadataTypeValidatorMetaclass
):
    _loader_module_name = 'metadata_validators'

    def validate(self, input_data):
        try:
            self.execute(input_data)
        except Exception as exception:
            raise ValidationError(message=exception)



from django.utils.translation import gettext_lazy as _

from mayan.apps.dynamic_search.search_models import SearchModel

from .permissions import permission_index_instance_view

search_model_index_instance_node = SearchModel(
    app_label='document_indexing', model_name='IndexInstanceNodeSearchResult',
    permission=permission_index_instance_view,
    serializer_path='mayan.apps.document_indexing.serializers.index_instance_serializers.IndexInstanceNodeSerializer'
)
search_model_index_instance_node.add_proxy_model(
    app_label='document_indexing', model_name='IndexInstanceNode'
)

search_model_index_instance_node.add_model_field(
    field='value', label=_(message='Value')
)

search_model_index_instance_node.add_model_field(
    field='documents__document_type__label', label=_(message='Document type')
)
search_model_index_instance_node.add_model_field(
    field='documents__files__mimetype', label=_(message='Document MIME type')
)
search_model_index_instance_node.add_model_field(
    field='documents__label', label=_(message='Document label')
)
search_model_index_instance_node.add_model_field(
    field='documents__description', label=_(message='Document description')
)
search_model_index_instance_node.add_model_field(
    field='documents__uuid', label=_(message='Document UUID')
)
search_model_index_instance_node.add_model_field(
    field='documents__files__checksum', label=_(message='Document checksum')
)



from django.apps import apps
from django.db.models.signals import post_save, pre_delete
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.acls.permissions import (
    permission_acl_edit, permission_acl_view
)
from mayan.apps.app_manager.apps import MayanAppConfig
from mayan.apps.common.classes import ModelCopy
from mayan.apps.common.menus import (
    menu_list_facet, menu_main, menu_multi_item, menu_object, menu_secondary
)
from mayan.apps.databases.classes import ModelQueryFields
from mayan.apps.events.classes import EventModelRegistry, ModelEventType
from mayan.apps.navigation.source_columns import SourceColumn
from mayan.apps.rest_api.fields import DynamicSerializerField

from .column_widgets import DocumentCabinetWidget
from .events import (
    event_cabinet_deleted, event_cabinet_document_added,
    event_cabinet_document_removed, event_cabinet_edited
)
from .handlers import handler_cabinet_pre_delete, handler_index_document
from .links import (
    link_cabinet_child_add, link_cabinet_create, link_cabinet_delete,
    link_cabinet_edit, link_cabinet_list, link_cabinet_view,
    link_custom_acl_list, link_document_cabinet_add,
    link_document_cabinet_list, link_document_cabinet_remove,
    link_document_multiple_cabinet_add, link_multiple_document_cabinet_remove
)
from .menus import menu_cabinets
from .methods import method_document_get_cabinets
from .permissions import (
    permission_cabinet_add_document, permission_cabinet_create,
    permission_cabinet_delete, permission_cabinet_edit,
    permission_cabinet_remove_document, permission_cabinet_view
)


class CabinetsApp(MayanAppConfig):
    app_namespace = 'cabinets'
    app_url = 'cabinets'
    has_rest_api = True
    has_static_media = True
    has_tests = True
    name = 'mayan.apps.cabinets'
    static_media_ignore_patterns = (
        'cabinets/node_modules/jstree/component.json',
        'cabinets/node_modules/jstree/jstree.jquery.json',
        'cabinets/node_modules/jstree/src/*',
    )
    verbose_name = _(message='Cabinets')

    def ready(self):
        super().ready()

        Cabinet = self.get_model(model_name='Cabinet')
        CabinetSearchResult = self.get_model(
            model_name='CabinetSearchResult'
        )
        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )
        DocumentCabinet = self.get_model(model_name='DocumentCabinet')
        DocumentFileSearchResult = apps.get_model(
            app_label='documents', model_name='DocumentFileSearchResult'
        )
        DocumentFilePageSearchResult = apps.get_model(
            app_label='documents', model_name='DocumentFilePageSearchResult'
        )
        DocumentVersionSearchResult = apps.get_model(
            app_label='documents', model_name='DocumentVersionSearchResult'
        )
        DocumentVersionPageSearchResult = apps.get_model(
            app_label='documents',
            model_name='DocumentVersionPageSearchResult'
        )

        # Add explicit order_by as DocumentCabinet ordering Meta option has no
        # effect.
        Document.add_to_class(
            name='get_cabinets', value=method_document_get_cabinets
        )

        DynamicSerializerField.add_serializer(
            klass=Cabinet,
            serializer_class='mayan.apps.cabinets.serializers.CabinetSerializer'
        )

        EventModelRegistry.register(model=Cabinet, acl_bind_link=False)

        def cabinet_model_copy_condition(instance):
            return instance.is_root_node()

        def cabinet_unique_conditional(instance, new_instance_dictionary):
            if instance.parent:
                return instance.parent_id == new_instance_dictionary['parent_id'] and instance.label == new_instance_dictionary['label']
            else:
                return not new_instance_dictionary['parent_id'] and instance.label == new_instance_dictionary['label']

        ModelCopy(
            model=Cabinet, condition=cabinet_model_copy_condition,
            bind_link=True, acl_bind_link=False, register_permission=True
        ).add_fields(
            field_names=('label', 'documents'), unique_conditional={
                'label': cabinet_unique_conditional
            }
        )

        ModelEventType.register(
            model=Cabinet, event_types=(
                event_cabinet_deleted, event_cabinet_edited,
                event_cabinet_document_added, event_cabinet_document_removed
            )
        )
        ModelEventType.register(
            model=Document, event_types=(
                event_cabinet_document_added, event_cabinet_document_removed
            )
        )

        ModelPermission.register(
            model=Cabinet, permissions=(
                permission_acl_edit, permission_acl_view,
                permission_cabinet_create, permission_cabinet_delete,
                permission_cabinet_edit, permission_cabinet_view,
                permission_cabinet_add_document,
                permission_cabinet_remove_document
            ), bind_link=False
        )

        ModelPermission.register(
            model=Document, permissions=(
                permission_cabinet_add_document,
                permission_cabinet_remove_document, permission_cabinet_view
            )
        )

        model_query_fields_document = ModelQueryFields(model=Document)
        model_query_fields_document.add_prefetch_related_field(
            field_name='cabinets'
        )

        def get_root_filter():
            return {
                'acl_filter': {'level': 0},
                'acl_values': ('tree_id',),
                'field_lookup': 'tree_id__in'
            }

        ModelPermission.register_field_query_function(
            model=Cabinet, function=get_root_filter
        )

        SourceColumn(
            attribute='label', is_identifier=True, is_sortable=True,
            source=Cabinet
        )

        SourceColumn(
            attribute='get_descendants_document_count', include_label=True,
            kwargs={'user': 'request.user'}, label=_(message='Documents'),
            source=Cabinet
        )

        SourceColumn(
            attribute='get_full_path', source=CabinetSearchResult
        )

        SourceColumn(
            label=_(message='Cabinets'), order=1, source=Document,
            widget=DocumentCabinetWidget
        )
        SourceColumn(
            attribute='document', label=_(message='Cabinets'), order=1,
            source=DocumentFileSearchResult, widget=DocumentCabinetWidget
        )
        SourceColumn(
            attribute='document_file__document', label=_(
                message='Cabinets'
            ), order=1, source=DocumentFilePageSearchResult,
            widget=DocumentCabinetWidget
        )
        SourceColumn(
            attribute='document', label=_(message='Cabinets'), order=1,
            source=DocumentVersionSearchResult, widget=DocumentCabinetWidget
        )
        SourceColumn(
            attribute='document_version__document', label=_(
                message='Cabinets'
            ), order=1, source=DocumentVersionPageSearchResult,
            widget=DocumentCabinetWidget
        )

        menu_list_facet.bind_links(
            links=(link_document_cabinet_list,), sources=(Document,)
        )

        menu_cabinets.bind_links(
            links=(
                link_cabinet_list, link_cabinet_create
            )
        )
        menu_list_facet.bind_links(
            links=(
                link_cabinet_view, link_custom_acl_list
            ),
            sources=(Cabinet,)
        )

        menu_main.bind_links(links=(menu_cabinets,), position=30)

        menu_multi_item.bind_links(
            links=(
                link_document_multiple_cabinet_add,
                link_multiple_document_cabinet_remove
            ), sources=(Document,)
        )
        menu_object.bind_links(
            exclude=(DocumentCabinet,),
            links=(
                link_cabinet_delete, link_cabinet_edit,
                link_cabinet_child_add
            ), sources=(Cabinet,)
        )
        menu_object.unbind_links(
            links=(
                link_cabinet_delete, link_cabinet_edit,
                link_cabinet_child_add
            ), sources=(DocumentCabinet,)
        )
        menu_secondary.bind_links(
            links=(link_document_cabinet_add, link_document_cabinet_remove),
            sources=(
                'cabinets:document_cabinet_list',
                'cabinets:document_cabinet_add',
                'cabinets:document_cabinet_remove'
            )
        )

        # Index update

        post_save.connect(
            dispatch_uid='cabinets_handler_index_document',
            receiver=handler_index_document,
            sender=Cabinet
        )
        pre_delete.connect(
            dispatch_uid='cabinets_handler_cabinet_pre_delete',
            receiver=handler_cabinet_pre_delete,
            sender=Cabinet
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.links import Link

from .icons import icon_app_list

link_app_list = Link(
    icon=icon_app_list, text=_(message='Apps'), view='app_manager:app_list'
)



from django.apps import apps

from mayan.apps.lock_manager.exceptions import LockError
from mayan.celery import app


@app.task(ignore_result=True)
def task_duplicates_clean_empty_lists():
    DuplicateBackendEntry = apps.get_model(
        app_label='duplicates', model_name='DuplicateBackendEntry'
    )
    DuplicateBackendEntry.objects.clean_empty_duplicate_lists()


@app.task(ignore_result=True)
def task_duplicates_scan_all():
    Document = apps.get_model(
        app_label='documents', model_name='Document'
    )

    for document in Document.valid.all():
        task_duplicates_scan_for.apply_async(
            kwargs={
                'document_id': document.pk
            }
        )


@app.task(bind=True, ignore_result=True)
def task_duplicates_scan_for(self, document_id):
    Document = apps.get_model(
        app_label='documents', model_name='Document'
    )
    StoredDuplicateBackend = apps.get_model(
        app_label='duplicates', model_name='StoredDuplicateBackend'
    )

    document = Document.objects.get(pk=document_id)

    try:
        StoredDuplicateBackend.objects.scan_document(
            document=document
        )
    except LockError as exception:
        raise self.retry(exc=exception)



from django.db import migrations


EVENT_NAME_MAP = {
    'file_metadata.document_version_submit': 'file_metadata.document_file_submitted',
    'file_metadata.document_version_finish': 'file_metadata.document_file_finished'
}


def code_rename_internal_event_name(apps, schema_editor):
    Action = apps.get_model(
        app_label='actstream', model_name='Action'
    )

    for key, value in EVENT_NAME_MAP.items():
        queryset = Action.objects.filter(verb=key)
        queryset.update(verb=value)


def reverse_code_rename_internal_event_name(apps, schema_editor):
    Action = apps.get_model(
        app_label='actstream', model_name='Action'
    )

    for key, value in EVENT_NAME_MAP.items():
        queryset = Action.objects.filter(verb=value)
        queryset.update(verb=key)


class Migration(migrations.Migration):
    dependencies = [
        ('file_metadata', '0017_documenttypedriverconfiguration_arguments')
    ]

    operations = [
        migrations.RunPython(
            code=code_rename_internal_event_name,
            reverse_code=reverse_code_rename_internal_event_name
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.serializers.document_type_serializers import (
    DocumentTypeSerializer
)
from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import MultiKwargHyperlinkedIdentityField

from ..models import DocumentTypeDriverConfiguration

from .file_metadata_serializers import StoredDriverSerializer


class DocumentTypeDriverConfigurationSerializer(serializers.ModelSerializer):
    document_type = DocumentTypeSerializer(
        label=_(message='Document type'), read_only=True
    )
    stored_driver = StoredDriverSerializer(
        label=_(message='Stored driver'), read_only=True
    )
    url = MultiKwargHyperlinkedIdentityField(
        label=_('URL'), view_kwargs=(
            {
                'lookup_field': 'document_type_id',
                'lookup_url_kwarg': 'document_type_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'driver_id'
            }
        ), view_name='rest_api:document_type_file_metadata_configuration-detail'
    )

    class Meta:
        fields = (
            'arguments', 'document_type', 'enabled', 'stored_driver', 'url'
        )
        model = DocumentTypeDriverConfiguration
        read_only_fields = ('document_type', 'stored_driver', 'url')



from django.contrib.auth.models import Group
from django.utils.translation import gettext_lazy as _

from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import FilteredPrimaryKeyRelatedField
from mayan.apps.user_management.permissions import permission_group_edit

from .classes import Permission
from .models import Role, StoredPermission


class PermissionSerializer(serializers.Serializer):
    namespace = serializers.CharField(
        label=_(message='Namespace'), read_only=True
    )
    pk = serializers.CharField(
        label=_(message='Primary key'), read_only=True
    )
    label = serializers.CharField(
        label=_(message='Label'), read_only=True
    )

    def to_representation(self, instance):
        if isinstance(instance, StoredPermission):
            return super().to_representation(
                instance.volatile_permission
            )
        else:
            return super().to_representation(
                instance
            )


class RoleGroupAddSerializer(serializers.Serializer):
    group = FilteredPrimaryKeyRelatedField(
        help_text=_(message='Primary key of the group to add to the role.'),
        label=_(message='Group ID'), source_queryset=Group.objects.all(),
        source_permission=permission_group_edit
    )


class RoleGroupRemoveSerializer(serializers.Serializer):
    group = FilteredPrimaryKeyRelatedField(
        help_text=_(message='Primary key of the group to remove from the role.'),
        label=_(message='Group ID'), source_queryset=Group.objects.all(),
        source_permission=permission_group_edit
    )


class RolePermissionAddSerializer(serializers.Serializer):
    permission = FilteredPrimaryKeyRelatedField(
        help_text=_(message='Primary key of the permission to add to the role.'),
        label=_(message='Permission ID'), source_queryset=Permission.all()
    )


class RolePermissionRemoveSerializer(serializers.Serializer):
    permission = FilteredPrimaryKeyRelatedField(
        help_text=_(
            message='Primary key of the permission to remove from the role.'
        ), label=_(message='Permission ID'), source_queryset=Permission.all()
    )


class RoleSerializer(serializers.HyperlinkedModelSerializer):
    groups_url = serializers.HyperlinkedIdentityField(
        label=_(message='Groups URL'), lookup_url_kwarg='role_id',
        view_name='rest_api:role-group-list'
    )
    groups_add_url = serializers.HyperlinkedIdentityField(
        label=_(message='Groups add URL'), lookup_url_kwarg='role_id',
        view_name='rest_api:role-group-add'
    )
    groups_remove_url = serializers.HyperlinkedIdentityField(
        label=_(message='Groups remove URL'), lookup_url_kwarg='role_id',
        view_name='rest_api:role-group-remove'
    )
    permissions_url = serializers.HyperlinkedIdentityField(
        label=_(message='Permissions URL'), lookup_url_kwarg='role_id',
        view_name='rest_api:role-permission-list'
    )
    permissions_add_url = serializers.HyperlinkedIdentityField(
        label=_(message='Permissions add URL'), lookup_url_kwarg='role_id',
        view_name='rest_api:role-permission-add'
    )
    permissions_remove_url = serializers.HyperlinkedIdentityField(
        label=_(message='Permissions remove URL'), lookup_url_kwarg='role_id',
        view_name='rest_api:role-permission-remove'
    )

    class Meta:
        extra_kwargs = {
            'url': {
                'label': _(message='URL'),
                'lookup_url_kwarg': 'role_id',
                'view_name': 'rest_api:role-detail'
            }
        }
        fields = (
            'groups_add_url', 'groups_remove_url', 'groups_url', 'id',
            'label', 'permissions_add_url', 'permissions_url',
            'permissions_remove_url', 'url'
        )
        model = Role



from mayan.apps.icons.icons import Icon

icon_user_locale_profile_detail = Icon(
    driver_name='fontawesome', symbol='globe'
)
icon_user_locale_profile_edit = Icon(
    driver_name='fontawesome-dual', primary_symbol='globe',
    secondary_symbol='pencil-alt'
)



from django.db import migrations, models

import mayan.apps.common.validators


class Migration(migrations.Migration):
    dependencies = [
        ('document_states', '0014_auto_20190701_0454')
    ]

    operations = [
        migrations.AddField(
            model_name='workflowtransitionfield',
            name='widget',
            field=models.PositiveIntegerField(
                blank=True, choices=[
                    (1, 'Text area')
                ],
                help_text='An optional class to change the default '
                'presentation of the field.', null=True,
                verbose_name='Widget class'
            )
        ),
        migrations.AddField(
            model_name='workflowtransitionfield',
            name='widget_kwargs',
            field=models.TextField(
                blank=True,
                help_text='A group of keyword arguments to customize the '
                'widget. Use YAML format.', validators=[
                    mayan.apps.common.validators.YAMLValidator()
                ], verbose_name='Widget keyword arguments'
            )
        ),
        migrations.AlterField(
            model_name='workflowinstance',
            name='context',
            field=models.TextField(blank=True, verbose_name='Context'),
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_b

queue_parsing = CeleryQueue(
    name='parsing', label=_(message='Parsing'), worker=worker_b
)

queue_parsing.add_task_type(
    dotted_path='mayan.apps.document_parsing.tasks.task_parse_document_file',
    label=_(message='Document file parsing')
)



from django.apps import apps

from mayan.apps.common.classes import PropertyHelper


class DocumentSourceMetadataHelper(PropertyHelper):
    @staticmethod
    @property
    def constructor(*args, **kwargs):
        return DocumentSourceMetadataHelper(*args, **kwargs)

    def get_result(self, name):
        file_latest = self.instance.file_latest

        if file_latest:
            return getattr(file_latest.source_metadata_value_of, name)
        else:
            return None


class DocumentFileSourceMetadataHelper(PropertyHelper):
    @staticmethod
    @property
    def constructor(*args, **kwargs):
        return DocumentFileSourceMetadataHelper(*args, **kwargs)

    def get_result(self, name):
        DocumentFileSourceMetadata = apps.get_model(
            app_label='sources', model_name='DocumentFileSourceMetadata'
        )

        try:
            source_metadata = self.instance.source_metadata.get(key=name)
        except DocumentFileSourceMetadata.DoesNotExist:
            return None
        else:
            return source_metadata.value



from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_(message='Document signatures'), name='document_signatures'
)

event_detached_signature_created = namespace.add_event_type(
    label=_(message='Detached signature created'), name='detached_signature_created'
)
event_detached_signature_deleted = namespace.add_event_type(
    label=_(message='Detached signature deleted'), name='detached_signature_deleted'
)
event_detached_signature_uploaded = namespace.add_event_type(
    label=_(message='Detached signature uploaded'),
    name='detached_signature_uploaded'
)
event_embedded_signature_created = namespace.add_event_type(
    label=_(message='Embedded signature created'), name='embedded_signature_created'
)



import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('file_caching', '0003_auto_20191130_2208')
    ]

    operations = [
        migrations.AlterField(
            model_name='cache',
            name='maximum_size',
            field=models.BigIntegerField(
                help_text='Maximum size of the cache in bytes.',
                validators=[
                    django.core.validators.MinValueValidator(limit_value=1)
                ], verbose_name='Maximum size'
            )
        )
    ]



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0090_alter_documentversion_active')
    ]

    operations = [
        migrations.AlterModelOptions(
            name='documenttype',
            options={
                'ordering': ('label',), 'verbose_name': 'Document type',
                'verbose_name_plural': 'Document types'
            }
        )
    ]



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('sources', '0028_auto_20210905_0558')
    ]

    operations = [
        migrations.RunPython(
            code=migrations.RunPython.noop,
            reverse_code=migrations.RunPython.noop
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.backends.forms import FormDynamicModelBackend
from mayan.apps.forms import form_fields, forms

from .classes import CredentialBackend
from .models import StoredCredential


class StoredCredentialBackendSelectionForm(forms.Form):
    backend = form_fields.ChoiceField(
        choices=(), help_text=_(
            message='The backend to use for the credential.'
        ), label=_(message='Backend')
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['backend'].choices = CredentialBackend.get_choices()


class StoredCredentialBackendDynamicForm(FormDynamicModelBackend):
    class Meta:
        fields = ('label', 'internal_name', 'backend_data')
        model = StoredCredential



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0016_auto_20150708_0325'),
    ]

    operations = [
        migrations.AlterField(
            model_name='document',
            name='date_added',
            field=models.DateTimeField(
                auto_now_add=True, verbose_name='Added', db_index=True
            ),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='documentversion',
            name='timestamp',
            field=models.DateTimeField(
                auto_now_add=True, verbose_name='Timestamp', db_index=True
            ),
            preserve_default=True,
        ),
    ]



from django.template.defaultfilters import filesizeformat
from django.utils.translation import gettext_lazy as _

from mayan.apps.forms import form_widgets, forms

from ..fields import DocumentFileField
from ..models.document_file_models import DocumentFile


class DocumentFileForm(forms.ModelForm):
    class Meta:
        fields = ('filename', 'comment',)
        model = DocumentFile


class DocumentFilePreviewForm(forms.Form):
    def __init__(self, *args, **kwargs):
        document_file = kwargs.pop('instance', None)
        transformation_instance_list = kwargs.pop(
            'transformation_instance_list', ()
        )
        super().__init__(*args, **kwargs)
        self.fields['document_file'].initial = document_file
        self.fields['document_file'].widget.attrs.update(
            {'transformation_instance_list': transformation_instance_list}
        )

    document_file = DocumentFileField()


class DocumentFilePropertiesForm(forms.DetailForm):
    """
    Detail class form to display a document file properties
    """
    fieldsets = (
        (
            _(message='Original'), {
                'fields': ('filename', 'size')
            }
        ), (
            _(message='Detected'), {
                'fields': ('mimetype', 'encoding', 'pages')
            }
        ), (
            _(message='Storage'), {
                'fields': ('file', 'exists', 'checksum')
            }
        ), (
            _(message='Other'), {
                'fields': ('timestamp', 'comment')
            }
        )
    )

    class Meta:
        extra_fields = [
            {
                'label': _(message='Date added'),
                'field': 'timestamp',
                'widget': form_widgets.DateTimeInput
            },
            {
                'label': _(message='MIME type'),
                'field': 'mimetype'
            },
            {
                'label': _(message='Encoding'),
                'field': 'encoding'
            },
            {
                'label': _(message='Size'),
                'func': lambda document_file: filesizeformat(
                    bytes_=document_file.size
                ) if document_file.size else '-', 'field': 'size'
            },
            {
                'label': _(message='Exists in storage'), 'field': 'exists'
            },
            {
                'label': _(message='Path in storage'), 'field': 'file'
            },
            {
                'label': _(message='Checksum'), 'field': 'checksum'
            },
            {
                'label': _(message='Pages'),
                'func': lambda document_file: document_file.pages.count(),
                'field': 'pages'
            }
        ]
        fields = ('filename', 'comment',)
        model = DocumentFile



from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.models.document_version_models import (
    DocumentVersion
)

from ..icons import (
    icon_document_version_attachment_send_single,
    icon_document_version_link_send_single
)

from .base import MailingObjectAttachmentSendView, MailingObjectLinkSendView


class MailDocumentVersionAttachmentView(MailingObjectAttachmentSendView):
    pk_url_kwarg = 'document_version_id'
    source_queryset = DocumentVersion.valid.all()
    success_message = _(
        message='%(count)d document version queued for email delivery'
    )
    success_message_plural = _(
        message='%(count)d document versions queued for email delivery'
    )
    title = 'Email document version'
    title_document = 'Email document version: %s'
    title_plural = 'Email documents version'
    view_icon = icon_document_version_attachment_send_single


class MailDocumentVersionLinkView(MailingObjectLinkSendView):
    pk_url_kwarg = 'document_version_id'
    source_queryset = DocumentVersion.valid.all()
    success_message = _(
        message='%(count)d document version link queued for email delivery'
    )
    success_message_plural = _(
        message='%(count)d document version links queued for email delivery'
    )
    title = 'Email document version link'
    title_document = 'Email link for document version: %s'
    title_plural = 'Email document version links'
    view_icon = icon_document_version_link_send_single



from functools import cache

from django.conf import settings
from django.contrib.admin.utils import (
    help_text_for_field, label_for_field, reverse_field_path
)
from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
from django.db.models.constants import LOOKUP_SEP

from .literals import DJANGO_SQLITE_BACKEND


def check_for_sqlite():
    return settings.DATABASES['default']['ENGINE'] == DJANGO_SQLITE_BACKEND and settings.DEBUG is False


def check_queryset(view, queryset):
    """
    Validate that a view queryset is usable.
    """
    try:
        queryset.query
    except AttributeError:
        # Check if it is an iterable.
        try:
            iter(queryset)
        except TypeError as exception:
            raise ImproperlyConfigured(
                'Queryset `{}` of view `{}` is not a valid queryset.'.format(
                    queryset, view.__class__
                )
            ) from exception
        else:
            return queryset
    else:
        return queryset


def instance_list_to_queryset(instance_list):
    manager = instance_list[0]._meta.default_manager

    return manager.filter(
        pk__in=[instance.pk for instance in instance_list]
    )


def _get_model_ordering_fields(
    model, exclude=None, prefix=None, separator='__'
):
    if prefix:
        prefix_string = '{}{}'.format(prefix, separator)
    else:
        prefix_string = ''

    model_ordering_fields = getattr(
        model, '_ordering_fields', ()
    )

    model_ordering_fields += (model._meta.pk.name,)

    for field_name in sorted(model_ordering_fields):
        yield '{}{}'.format(prefix_string, field_name)

    for field in model._meta.fields:
        if field.concrete:
            related_model, path = reverse_field_path(
                model=model, path=field.name
            )

            if path and related_model != exclude:
                prefix_related = '{}{}'.format(prefix_string, field.name)

                yield from _get_model_ordering_fields(
                    exclude=model, model=related_model, prefix=prefix_related
                )


@cache
def help_text_for_field_recursive(model, name):
    name, model = get_model_attribute_recursive(attribute=name, model=model)
    help_text = help_text_for_field(model=model, name=name)
    return help_text


@cache
def label_for_field_recursive(model, name):
    name, model = get_model_attribute_recursive(attribute=name, model=model)
    label = label_for_field(model=model, name=name)
    return label


@cache
def get_model_attribute_recursive(attribute, model):
    """
    Walk over the double underscore (__) separated path to the last
    field. Returns the field name and the corresponding model class.
    Used to introspect the label or short_description of a model's
    attribute.
    """
    last_model = model

    for part in attribute.split(LOOKUP_SEP):
        last_model = model
        try:
            field = model._meta.get_field(field_name=part)
        except FieldDoesNotExist:
            break
        else:
            model = field.related_model or field.model

    return (part, last_model)


@cache
def get_model_ordering_fields(model):
    iterable_field_list = _get_model_ordering_fields(model=model)
    field_list = tuple(iterable_field_list)
    return sorted(field_list)



from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.http import StreamingHttpResponse
from django.views.decorators.cache import patch_cache_control

from rest_framework.exceptions import APIException

from mayan.apps.mime_types.classes import MIMETypeBackend

from .classes import AppImageErrorImage, ConverterBase
from .exceptions import AppImageError
from .settings import (
    setting_image_cache_time, setting_image_generation_timeout
)
from .tasks import task_content_object_image_generate
from .utils import IndexedDictionary, factory_file_generator


class APIImageViewMixin:
    """
    get: Returns an image representation of the selected object.
    """
    def get_content_type(self):
        return ContentType.objects.get_for_model(model=self.obj)

    def get_file_generator(self):
        return factory_file_generator(image_object=self.cache_file)

    def get_serializer(self, *args, **kwargs):
        return None

    def get_serializer_class(self):
        return None

    def get_stream_mime_type(self):
        mime_type_backend = MIMETypeBackend.get_backend_instance()
        with self.cache_file.open() as file_object:
            mime_type, mime_encoding = mime_type_backend.get_mime_type(
                file_object=file_object, mime_type_only=True
            )
            return mime_type

    def retrieve(self, request, **kwargs):
        self.set_object()

        try:
            self.set_cache_file(request=request)
        except AppImageError as exception:
            app_image_error_image = AppImageErrorImage.get(
                name=exception.error_name
            )

            error_image_template_result = app_image_error_image.render(
                context={'details': exception.details}
            )

            detail = {
                'app_image_error_image_template': error_image_template_result
            }

            if exception.details:
                detail['details'] = exception.details

            raise APIException(detail=detail)
        else:
            file_generator = self.get_file_generator()

            content_type = ConverterBase.get_output_content_type()

            if not content_type:
                content_type = self.get_stream_mime_type()

            response = StreamingHttpResponse(
                content_type=content_type, streaming_content=file_generator()
            )

            if '_hash' in request.GET:
                patch_cache_control(
                    max_age=setting_image_cache_time.value,
                    response=response
                )
            return response

    def set_cache_file(self, request):
        query_dict = request.GET

        transformation_dictionary_list = IndexedDictionary(
            dictionary=query_dict
        ).as_dictionary_list()

        # An empty string is not a valid value for maximum_layer_order.
        # Fallback to None in case of a empty string.
        maximum_layer_order = request.GET.get('maximum_layer_order') or None
        if maximum_layer_order:
            maximum_layer_order = int(maximum_layer_order)

        task = task_content_object_image_generate.apply_async(
            kwargs={
                'content_type_id': self.get_content_type().pk,
                'object_id': self.obj.pk,
                'maximum_layer_order': maximum_layer_order,
                'transformation_dictionary_list': transformation_dictionary_list,
                'user_id': request.user.pk
            }
        )

        kwargs = {'timeout': setting_image_generation_timeout.value}
        if settings.DEBUG:
            # In debug more, task are run synchronously, causing this method
            # to be called inside another task. Disable the check of nested
            # tasks when using debug mode.
            kwargs['disable_sync_subtasks'] = False

        cache_filename = task.get(**kwargs)

        self.cache_file = self.obj.cache_partition.get_file(
            filename=cache_filename
        )

    def set_object(self):
        self.obj = self.get_object()
